# Copyright (c) 2019-2021, Manfred Moitzi
# License: MIT License
from functools import partial
from ezdxf.lldxf.hdrvars import SingleValue, Point2D, Point3D, HeaderVarDef
from ezdxf.lldxf.const import (
    DXF12,
    DXF2000,
    DXF2004,
    DXF2007,
    DXF2010,
    DXF2013,
    DXF2018,
)

CONST_GUID = "{00000000-0000-0000-0000-000000000000}"

HEADER_VAR_MAP = {
    "$ACADVER": HeaderVarDef(
        name="$ACADVER",
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=0,
        default="AC1032",
    ),
    "$ACADMAINTVER": HeaderVarDef(
        name="$ACADMAINTVER",
        # group code changed to 90 in DXF R2018+, this fact is handled in:
        # ezdxf.sections.header.HeaderSection.export_dxf()
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=100,
        default=4,
    ),
    "$DWGCODEPAGE": HeaderVarDef(
        name="$DWGCODEPAGE",
        code=3,
        factory=partial(SingleValue, code=3),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=200,
        default="ANSI_1252",
    ),
    "$LASTSAVEDBY": HeaderVarDef(
        name="$LASTSAVEDBY",
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=300,
        default="ezdxf",
    ),
    "$REQUIREDVERSIONS": HeaderVarDef(
        name="$REQUIREDVERSIONS",
        code=160,
        factory=partial(SingleValue, code=160),
        mindxf=DXF2013,
        maxdxf=DXF2018,
        priority=400,
        default=0,
    ),
    "$INSBASE": HeaderVarDef(
        name="$INSBASE",
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=500,
        default=(0.0, 0.0, 0.0),
    ),
    "$EXTMIN": HeaderVarDef(
        name="$EXTMIN",
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=600,
        default=(1e20, 1e20, 1e20),
    ),
    "$EXTMAX": HeaderVarDef(
        name="$EXTMAX",
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=700,
        default=(-1e20, -1e20, -1e20),
    ),
    "$LIMMIN": HeaderVarDef(
        name="$LIMMIN",
        code=10,
        factory=Point2D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=800,
        default=(0.0, 0.0),
    ),
    "$LIMMAX": HeaderVarDef(
        name="$LIMMAX",
        code=10,
        factory=Point2D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=900,
        default=(420.0, 297.0),
    ),
    "$ORTHOMODE": HeaderVarDef(
        name="$ORTHOMODE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1000,
        default=0,
    ),
    "$REGENMODE": HeaderVarDef(
        name="$REGENMODE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1100,
        default=1,
    ),
    "$FILLMODE": HeaderVarDef(
        name="$FILLMODE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1200,
        default=1,
    ),
    "$DRAGMODE": HeaderVarDef(
        name="$DRAGMODE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF12,
        priority=1250,
        default=2,
    ),
    "$QTEXTMODE": HeaderVarDef(
        name="$QTEXTMODE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1300,
        default=0,
    ),
    "$MIRRTEXT": HeaderVarDef(
        name="$MIRRTEXT",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1400,
        default=1,
    ),
    "$OSMODE": HeaderVarDef(
        name="$OSMODE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF12,
        priority=1400,
        default=20583,
    ),
    "$LTSCALE": HeaderVarDef(
        name="$LTSCALE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1500,
        default=1.0,
    ),
    "$ATTMODE": HeaderVarDef(
        name="$ATTMODE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1600,
        default=1,
    ),
    "$TEXTSIZE": HeaderVarDef(
        name="$TEXTSIZE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1700,
        default=2.5,
    ),
    "$TRACEWID": HeaderVarDef(
        name="$TRACEWID",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1800,
        default=1.0,
    ),
    "$TEXTSTYLE": HeaderVarDef(
        name="$TEXTSTYLE",
        code=7,
        factory=partial(SingleValue, code=7),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=1900,
        default="Standard",
    ),
    "$CLAYER": HeaderVarDef(
        name="$CLAYER",
        code=8,
        factory=partial(SingleValue, code=8),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2000,
        default="0",
    ),
    "$CELTYPE": HeaderVarDef(
        name="$CELTYPE",
        code=6,
        factory=partial(SingleValue, code=6),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2100,
        default="ByLayer",
    ),
    "$CECOLOR": HeaderVarDef(
        name="$CECOLOR",
        code=62,
        factory=partial(SingleValue, code=62),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2200,
        default=256,
    ),
    "$CELTSCALE": HeaderVarDef(
        name="$CELTSCALE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=2300,
        default=1.0,
    ),
    "$DISPSILH": HeaderVarDef(
        name="$DISPSILH",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=2400,
        default=0,
    ),
    "$DIMSCALE": HeaderVarDef(
        name="$DIMSCALE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2500,
        default=1.0,
    ),
    "$DIMASZ": HeaderVarDef(
        name="$DIMASZ",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2600,
        default=2.5,
    ),
    "$DIMEXO": HeaderVarDef(
        name="$DIMEXO",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2700,
        default=0.625,
    ),
    "$DIMDLI": HeaderVarDef(
        name="$DIMDLI",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2800,
        default=3.75,
    ),
    "$DIMRND": HeaderVarDef(
        name="$DIMRND",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=2900,
        default=0.0,
    ),
    "$DIMDLE": HeaderVarDef(
        name="$DIMDLE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3000,
        default=0.0,
    ),
    "$DIMEXE": HeaderVarDef(
        name="$DIMEXE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3100,
        default=1.25,
    ),
    "$DIMTP": HeaderVarDef(
        name="$DIMTP",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3200,
        default=0.0,
    ),
    "$DIMTM": HeaderVarDef(
        name="$DIMTM",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3300,
        default=0.0,
    ),
    "$DIMTXT": HeaderVarDef(
        name="$DIMTXT",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3400,
        default=2.5,
    ),
    "$DIMCEN": HeaderVarDef(
        name="$DIMCEN",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3500,
        default=2.5,
    ),
    "$DIMTSZ": HeaderVarDef(
        name="$DIMTSZ",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3600,
        default=0.0,
    ),
    "$DIMTOL": HeaderVarDef(
        name="$DIMTOL",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3700,
        default=0,
    ),
    "$DIMLIM": HeaderVarDef(
        name="$DIMLIM",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3800,
        default=0,
    ),
    "$DIMTIH": HeaderVarDef(
        name="$DIMTIH",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=3900,
        default=0,
    ),
    "$DIMTOH": HeaderVarDef(
        name="$DIMTOH",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4000,
        default=0,
    ),
    "$DIMSE1": HeaderVarDef(
        name="$DIMSE1",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4100,
        default=0,
    ),
    "$DIMSE2": HeaderVarDef(
        name="$DIMSE2",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4200,
        default=0,
    ),
    "$DIMTAD": HeaderVarDef(
        name="$DIMTAD",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4300,
        default=1,
    ),
    "$DIMZIN": HeaderVarDef(
        name="$DIMZIN",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4400,
        default=8,
    ),
    "$DIMBLK": HeaderVarDef(
        name="$DIMBLK",
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4500,
        default="",
    ),
    "$DIMASO": HeaderVarDef(
        name="$DIMASO",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4600,
        default=1,
    ),
    "$DIMSHO": HeaderVarDef(
        name="$DIMSHO",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4700,
        default=1,
    ),
    "$DIMPOST": HeaderVarDef(
        name="$DIMPOST",
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4800,
        default="",
    ),
    "$DIMAPOST": HeaderVarDef(
        name="$DIMAPOST",
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=4900,
        default="",
    ),
    "$DIMALT": HeaderVarDef(
        name="$DIMALT",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5000,
        default=0,
    ),
    "$DIMALTD": HeaderVarDef(
        name="$DIMALTD",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5100,
        default=3,
    ),
    "$DIMALTF": HeaderVarDef(
        name="$DIMALTF",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5200,
        default=0.03937007874,
    ),
    "$DIMLFAC": HeaderVarDef(
        name="$DIMLFAC",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5300,
        default=1.0,
    ),
    "$DIMTOFL": HeaderVarDef(
        name="$DIMTOFL",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5400,
        default=1,
    ),
    "$DIMTVP": HeaderVarDef(
        name="$DIMTVP",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5500,
        default=0.0,
    ),
    "$DIMTIX": HeaderVarDef(
        name="$DIMTIX",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5600,
        default=0,
    ),
    "$DIMSOXD": HeaderVarDef(
        name="$DIMSOXD",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5700,
        default=0,
    ),
    "$DIMSAH": HeaderVarDef(
        name="$DIMSAH",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5800,
        default=0,
    ),
    "$DIMBLK1": HeaderVarDef(
        name="$DIMBLK1",
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=5900,
        default="",
    ),
    "$DIMBLK2": HeaderVarDef(
        name="$DIMBLK2",
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=6000,
        default="",
    ),
    "$DIMSTYLE": HeaderVarDef(
        name="$DIMSTYLE",
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=6100,
        default="ISO-25",
    ),
    "$DIMCLRD": HeaderVarDef(
        name="$DIMCLRD",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=6200,
        default=0,
    ),
    "$DIMCLRE": HeaderVarDef(
        name="$DIMCLRE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=6300,
        default=0,
    ),
    "$DIMCLRT": HeaderVarDef(
        name="$DIMCLRT",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=6400,
        default=0,
    ),
    "$DIMTFAC": HeaderVarDef(
        name="$DIMTFAC",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=6500,
        default=1.0,
    ),
    "$DIMGAP": HeaderVarDef(
        name="$DIMGAP",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=6600,
        default=0.625,
    ),
    "$DIMJUST": HeaderVarDef(
        name="$DIMJUST",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=6700,
        default=0,
    ),
    "$DIMSD1": HeaderVarDef(
        name="$DIMSD1",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=6800,
        default=0,
    ),
    "$DIMSD2": HeaderVarDef(
        name="$DIMSD2",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=6900,
        default=0,
    ),
    "$DIMTOLJ": HeaderVarDef(
        name="$DIMTOLJ",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7000,
        default=0,
    ),
    "$DIMTZIN": HeaderVarDef(
        name="$DIMTZIN",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7100,
        default=8,
    ),
    "$DIMALTZ": HeaderVarDef(
        name="$DIMALTZ",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7200,
        default=0,
    ),
    "$DIMALTTZ": HeaderVarDef(
        name="$DIMALTTZ",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7300,
        default=0,
    ),
    "$DIMUPT": HeaderVarDef(
        name="$DIMUPT",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7400,
        default=0,
    ),
    "$DIMDEC": HeaderVarDef(
        name="$DIMDEC",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7500,
        default=2,
    ),
    "$DIMTDEC": HeaderVarDef(
        name="$DIMTDEC",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7600,
        default=2,
    ),
    "$DIMALTU": HeaderVarDef(
        name="$DIMALTU",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7700,
        default=2,
    ),
    "$DIMALTTD": HeaderVarDef(
        name="$DIMALTTD",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7800,
        default=3,
    ),
    "$DIMTXSTY": HeaderVarDef(
        name="$DIMTXSTY",
        code=7,
        factory=partial(SingleValue, code=7),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=7900,
        default="Standard",
    ),
    "$DIMAUNIT": HeaderVarDef(
        name="$DIMAUNIT",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8000,
        default=0,
    ),
    "$DIMADEC": HeaderVarDef(
        name="$DIMADEC",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8100,
        default=0,
    ),
    "$DIMALTRND": HeaderVarDef(
        name="$DIMALTRND",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8200,
        default=0.0,
    ),
    "$DIMAZIN": HeaderVarDef(
        name="$DIMAZIN",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8300,
        default=0,
    ),
    "$DIMDSEP": HeaderVarDef(
        name="$DIMDSEP",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8400,
        default=44,
    ),
    "$DIMATFIT": HeaderVarDef(
        name="$DIMATFIT",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8500,
        default=3,
    ),
    "$DIMFRAC": HeaderVarDef(
        name="$DIMFRAC",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8600,
        default=0,
    ),
    "$DIMLDRBLK": HeaderVarDef(
        name="$DIMLDRBLK",
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8700,
        default="",
    ),
    "$DIMLUNIT": HeaderVarDef(
        name="$DIMLUNIT",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8800,
        default=2,
    ),
    "$COORDS": HeaderVarDef(
        name="$COORDS",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF12,
        priority=8800,
        default=1,
    ),
    "$DIMLWD": HeaderVarDef(
        name="$DIMLWD",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=8900,
        default=-2,
    ),
    "$DIMLWE": HeaderVarDef(
        name="$DIMLWE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=9000,
        default=-2,
    ),
    "$DIMTMOVE": HeaderVarDef(
        name="$DIMTMOVE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=9100,
        default=0,
    ),
    "$DIMFXL": HeaderVarDef(
        name="$DIMFXL",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9200,
        default=1.0,
    ),
    "$ATTDIA": HeaderVarDef(
        name="$ATTDIA",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF12,
        priority=9200,
        default=0,
    ),
    "$DIMFXLON": HeaderVarDef(
        name="$DIMFXLON",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9300,
        default=0,
    ),
    "$ATTREQ": HeaderVarDef(
        name="$ATTREQ",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF12,
        priority=9300,
        default=1,
    ),
    "$DIMJOGANG": HeaderVarDef(
        name="$DIMJOGANG",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9400,
        default=0.785398163397,
    ),
    "$DIMTFILL": HeaderVarDef(
        name="$DIMTFILL",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9500,
        default=0,
    ),
    "$DIMTFILLCLR": HeaderVarDef(
        name="$DIMTFILLCLR",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9600,
        default=0,
    ),
    "$DIMARCSYM": HeaderVarDef(
        name="$DIMARCSYM",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9700,
        default=0,
    ),
    "$DIMLTYPE": HeaderVarDef(
        name="$DIMLTYPE",
        code=6,
        factory=partial(SingleValue, code=6),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9800,
        default="",
    ),
    "$DIMLTEX1": HeaderVarDef(
        name="$DIMLTEX1",
        code=6,
        factory=partial(SingleValue, code=6),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=9900,
        default="",
    ),
    "$DIMLTEX2": HeaderVarDef(
        name="$DIMLTEX2",
        code=6,
        factory=partial(SingleValue, code=6),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=10000,
        default="",
    ),
    "$DIMTXTDIRECTION": HeaderVarDef(
        name="$DIMTXTDIRECTION",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2010,
        maxdxf=DXF2018,
        priority=10100,
        default=0,
    ),
    "$LUNITS": HeaderVarDef(
        name="$LUNITS",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10200,
        default=2,
    ),
    "$LUPREC": HeaderVarDef(
        name="$LUPREC",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10300,
        default=4,
    ),
    "$SKETCHINC": HeaderVarDef(
        name="$SKETCHINC",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10400,
        default=1.0,
    ),
    "$FILLETRAD": HeaderVarDef(
        name="$FILLETRAD",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10500,
        default=10.0,
    ),
    "$AUNITS": HeaderVarDef(
        name="$AUNITS",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10600,
        default=0,
    ),
    "$AUPREC": HeaderVarDef(
        name="$AUPREC",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10700,
        default=2,
    ),
    "$MENU": HeaderVarDef(
        name="$MENU",
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10800,
        default=".",
    ),
    "$ELEVATION": HeaderVarDef(
        name="$ELEVATION",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=10900,
        default=0.0,
    ),
    "$PELEVATION": HeaderVarDef(
        name="$PELEVATION",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=11000,
        default=0.0,
    ),
    "$THICKNESS": HeaderVarDef(
        name="$THICKNESS",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=11100,
        default=0.0,
    ),
    "$LIMCHECK": HeaderVarDef(
        name="$LIMCHECK",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=11200,
        default=0,
    ),
    "$CHAMFERA": HeaderVarDef(
        name="$CHAMFERA",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=11300,
        default=0.0,
    ),
    "$CHAMFERB": HeaderVarDef(
        name="$CHAMFERB",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=11400,
        default=0.0,
    ),
    "$CHAMFERC": HeaderVarDef(
        name="$CHAMFERC",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=11500,
        default=0.0,
    ),
    "$CHAMFERD": HeaderVarDef(
        name="$CHAMFERD",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=11600,
        default=0.0,
    ),
    "$SKPOLY": HeaderVarDef(
        name="$SKPOLY",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=11700,
        default=0,
    ),
    "$TDCREATE": HeaderVarDef(
        name="$TDCREATE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=11800,
        default=2458532.195663565,
    ),
    "$TDUCREATE": HeaderVarDef(
        name="$TDUCREATE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=11900,
        default=2458532.153996898,
    ),
    "$TDUPDATE": HeaderVarDef(
        name="$TDUPDATE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12000,
        default=2458532.196097766,
    ),
    "$TDUUPDATE": HeaderVarDef(
        name="$TDUUPDATE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=12100,
        default=2458532.1544311,
    ),
    "$TDINDWG": HeaderVarDef(
        name="$TDINDWG",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12200,
        default=0.0,
    ),
    "$TDUSRTIMER": HeaderVarDef(
        name="$TDUSRTIMER",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12300,
        default=0.0,
    ),
    "$USRTIMER": HeaderVarDef(
        name="$USRTIMER",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12400,
        default=1,
    ),
    "$ANGBASE": HeaderVarDef(
        name="$ANGBASE",
        code=50,
        factory=partial(SingleValue, code=50),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12500,
        default=0.0,
    ),
    "$ANGDIR": HeaderVarDef(
        name="$ANGDIR",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12600,
        default=0,
    ),
    "$PDMODE": HeaderVarDef(
        name="$PDMODE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12700,
        default=0,
    ),
    "$PDSIZE": HeaderVarDef(
        name="$PDSIZE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12800,
        default=0.0,
    ),
    "$PLINEWID": HeaderVarDef(
        name="$PLINEWID",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=12900,
        default=0.0,
    ),
    "$SPLFRAME": HeaderVarDef(
        name="$SPLFRAME",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13000,
        default=0,
    ),
    "$SPLINETYPE": HeaderVarDef(
        name="$SPLINETYPE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13100,
        default=6,
    ),
    "$SPLINESEGS": HeaderVarDef(
        name="$SPLINESEGS",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13200,
        default=8,
    ),
    "$HANDLING": HeaderVarDef(
        name="$HANDLING",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF12,
        priority=9400,
        default=1,
    ),
    "$HANDSEED": HeaderVarDef(
        name="$HANDSEED",
        code=5,
        factory=partial(SingleValue, code=5),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13300,
        default="100",
    ),
    "$SURFTAB1": HeaderVarDef(
        name="$SURFTAB1",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13400,
        default=6,
    ),
    "$SURFTAB2": HeaderVarDef(
        name="$SURFTAB2",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13500,
        default=6,
    ),
    "$SURFTYPE": HeaderVarDef(
        name="$SURFTYPE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13600,
        default=6,
    ),
    "$SURFU": HeaderVarDef(
        name="$SURFU",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13700,
        default=6,
    ),
    "$SURFV": HeaderVarDef(
        name="$SURFV",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=13800,
        default=6,
    ),
    "$UCSBASE": HeaderVarDef(
        name="$UCSBASE",
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=13900,
        default="",
    ),
    "$UCSNAME": HeaderVarDef(
        name="$UCSNAME",
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=14000,
        default="",
    ),
    "$UCSORG": HeaderVarDef(
        name="$UCSORG",
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=14100,
        default=(0.0, 0.0, 0.0),
    ),
    "$UCSXDIR": HeaderVarDef(
        name="$UCSXDIR",
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=14200,
        default=(1.0, 0.0, 0.0),
    ),
    "$UCSYDIR": HeaderVarDef(
        name="$UCSYDIR",
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=14300,
        default=(0.0, 1.0, 0.0),
    ),
    "$UCSORTHOREF": HeaderVarDef(
        name="$UCSORTHOREF",
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=14400,
        default="",
    ),
    "$UCSORTHOVIEW": HeaderVarDef(
        name="$UCSORTHOVIEW",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=14500,
        default=0,
    ),
    "$UCSORGTOP": HeaderVarDef(
        name="$UCSORGTOP",
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=14600,
        default=(0.0, 0.0, 0.0),
    ),
    "$UCSORGBOTTOM": HeaderVarDef(
        name="$UCSORGBOTTOM",
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=14700,
        default=(0.0, 0.0, 0.0),
    ),
    "$UCSORGLEFT": HeaderVarDef(
        name="$UCSORGLEFT",
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=14800,
        default=(0.0, 0.0, 0.0),
    ),
    "$UCSORGRIGHT": HeaderVarDef(
        name="$UCSORGRIGHT",
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=14900,
        default=(0.0, 0.0, 0.0),
    ),
    "$UCSORGFRONT": HeaderVarDef(
        name="$UCSORGFRONT",
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=15000,
        default=(0.0, 0.0, 0.0),
    ),
    "$UCSORGBACK": HeaderVarDef(
        name="$UCSORGBACK",
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=15100,
        default=(0.0, 0.0, 0.0),
    ),
    "$PUCSBASE": HeaderVarDef(
        name="$PUCSBASE",
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=15200,
        default="",
    ),
    "$PUCSNAME": HeaderVarDef(
        name="$PUCSNAME",
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=15300,
        default="",
    ),
    "$PUCSORG": HeaderVarDef(
        name="$PUCSORG",
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=15400,
        default=(0.0, 0.0, 0.0),
    ),
    "$PUCSXDIR": HeaderVarDef(
        name="$PUCSXDIR",
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=15500,
        default=(1.0, 0.0, 0.0),
    ),
    "$PUCSYDIR": HeaderVarDef(
        name="$PUCSYDIR",
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=15600,
        default=(0.0, 1.0, 0.0),
    ),
    "$PUCSORTHOREF": HeaderVarDef(
        name="$PUCSORTHOREF",
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=15700,
        default="",
    ),
    "$PUCSORTHOVIEW": HeaderVarDef(
        name="$PUCSORTHOVIEW",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=15800,
        default=0,
    ),
    "$PUCSORGTOP": HeaderVarDef(
        name="$PUCSORGTOP",
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=15900,
        default=(0.0, 0.0, 0.0),
    ),
    "$PUCSORGBOTTOM": HeaderVarDef(
        name="$PUCSORGBOTTOM",
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=16000,
        default=(0.0, 0.0, 0.0),
    ),
    "$PUCSORGLEFT": HeaderVarDef(
        name="$PUCSORGLEFT",
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=16100,
        default=(0.0, 0.0, 0.0),
    ),
    "$PUCSORGRIGHT": HeaderVarDef(
        name="$PUCSORGRIGHT",
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=16200,
        default=(0.0, 0.0, 0.0),
    ),
    "$PUCSORGFRONT": HeaderVarDef(
        name="$PUCSORGFRONT",
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=16300,
        default=(0.0, 0.0, 0.0),
    ),
    "$PUCSORGBACK": HeaderVarDef(
        name="$PUCSORGBACK",
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=16400,
        default=(0.0, 0.0, 0.0),
    ),
    "$USERI1": HeaderVarDef(
        name="$USERI1",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=16500,
        default=0,
    ),
    "$USERI2": HeaderVarDef(
        name="$USERI2",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=16600,
        default=0,
    ),
    "$USERI3": HeaderVarDef(
        name="$USERI3",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=16700,
        default=0,
    ),
    "$USERI4": HeaderVarDef(
        name="$USERI4",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=16800,
        default=0,
    ),
    "$USERI5": HeaderVarDef(
        name="$USERI5",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=16900,
        default=0,
    ),
    "$USERR1": HeaderVarDef(
        name="$USERR1",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17000,
        default=0.0,
    ),
    "$USERR2": HeaderVarDef(
        name="$USERR2",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17100,
        default=0.0,
    ),
    "$USERR3": HeaderVarDef(
        name="$USERR3",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17200,
        default=0.0,
    ),
    "$USERR4": HeaderVarDef(
        name="$USERR4",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17300,
        default=0.0,
    ),
    "$USERR5": HeaderVarDef(
        name="$USERR5",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17400,
        default=0.0,
    ),
    "$WORLDVIEW": HeaderVarDef(
        name="$WORLDVIEW",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17500,
        default=1,
    ),
    "$SHADEDGE": HeaderVarDef(
        name="$SHADEDGE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17600,
        default=3,
    ),
    "$SHADEDIF": HeaderVarDef(
        name="$SHADEDIF",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17700,
        default=70,
    ),
    "$TILEMODE": HeaderVarDef(
        name="$TILEMODE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17800,
        default=1,
    ),
    "$MAXACTVP": HeaderVarDef(
        name="$MAXACTVP",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=17900,
        default=64,
    ),
    "$PINSBASE": HeaderVarDef(
        name="$PINSBASE",
        code=10,
        factory=Point3D,
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=18000,
        default=(0.0, 0.0, 0.0),
    ),
    "$PLIMCHECK": HeaderVarDef(
        name="$PLIMCHECK",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18100,
        default=0,
    ),
    "$PEXTMIN": HeaderVarDef(
        name="$PEXTMIN",
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18200,
        default=(1e20, 1e20, 1e20),
    ),
    "$PEXTMAX": HeaderVarDef(
        name="$PEXTMAX",
        code=10,
        factory=Point3D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18300,
        default=(-1e20, -1e20, -1e20),
    ),
    "$PLIMMIN": HeaderVarDef(
        name="$PLIMMIN",
        code=10,
        factory=Point2D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18400,
        default=(0.0, 0.0),
    ),
    "$PLIMMAX": HeaderVarDef(
        name="$PLIMMAX",
        code=10,
        factory=Point2D,
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18500,
        default=(420.0, 297.0),
    ),
    "$UNITMODE": HeaderVarDef(
        name="$UNITMODE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18600,
        default=0,
    ),
    "$VISRETAIN": HeaderVarDef(
        name="$VISRETAIN",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18700,
        default=1,
    ),
    "$PLINEGEN": HeaderVarDef(
        name="$PLINEGEN",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18800,
        default=0,
    ),
    "$PSLTSCALE": HeaderVarDef(
        name="$PSLTSCALE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF12,
        maxdxf=DXF2018,
        priority=18900,
        default=1,
    ),
    "$TREEDEPTH": HeaderVarDef(
        name="$TREEDEPTH",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19000,
        default=3020,
    ),
    "$CMLSTYLE": HeaderVarDef(
        name="$CMLSTYLE",
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19100,
        default="Standard",
    ),
    "$CMLJUST": HeaderVarDef(
        name="$CMLJUST",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19200,
        default=0,
    ),
    "$CMLSCALE": HeaderVarDef(
        name="$CMLSCALE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19300,
        default=20.0,
    ),
    "$PROXYGRAPHICS": HeaderVarDef(
        name="$PROXYGRAPHICS",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19400,
        default=1,
    ),
    "$MEASUREMENT": HeaderVarDef(
        name="$MEASUREMENT",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19500,
        default=1,
    ),
    "$CELWEIGHT": HeaderVarDef(
        name="$CELWEIGHT",
        code=370,
        factory=partial(SingleValue, code=370),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19600,
        default=-1,
    ),
    "$ENDCAPS": HeaderVarDef(
        name="$ENDCAPS",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19700,
        default=0,
    ),
    "$JOINSTYLE": HeaderVarDef(
        name="$JOINSTYLE",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19800,
        default=0,
    ),
    "$LWDISPLAY": HeaderVarDef(
        name="$LWDISPLAY",
        code=290,
        factory=partial(SingleValue, code=290),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=19900,
        default=0,
    ),
    "$INSUNITS": HeaderVarDef(
        name="$INSUNITS",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20000,
        default=6,
    ),
    "$HYPERLINKBASE": HeaderVarDef(
        name="$HYPERLINKBASE",
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20100,
        default="",
    ),
    "$STYLESHEET": HeaderVarDef(
        name="$STYLESHEET",
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20200,
        default="",
    ),
    "$XEDIT": HeaderVarDef(
        name="$XEDIT",
        code=290,
        factory=partial(SingleValue, code=290),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20300,
        default=1,
    ),
    "$CEPSNTYPE": HeaderVarDef(
        name="$CEPSNTYPE",
        code=380,
        factory=partial(SingleValue, code=380),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20400,
        default=0,
    ),
    "$PSTYLEMODE": HeaderVarDef(
        name="$PSTYLEMODE",
        code=290,
        factory=partial(SingleValue, code=290),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20500,
        default=1,
    ),
    "$FINGERPRINTGUID": HeaderVarDef(
        name="$FINGERPRINTGUID",
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20600,
        default=CONST_GUID,
    ),
    "$VERSIONGUID": HeaderVarDef(
        name="$VERSIONGUID",
        code=2,
        factory=partial(SingleValue, code=2),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20700,
        default=CONST_GUID,
    ),
    "$EXTNAMES": HeaderVarDef(
        name="$EXTNAMES",
        code=290,
        factory=partial(SingleValue, code=290),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20800,
        default=1,
    ),
    "$PSVPSCALE": HeaderVarDef(
        name="$PSVPSCALE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=20900,
        default=0.0,
    ),
    "$OLESTARTUP": HeaderVarDef(
        name="$OLESTARTUP",
        code=290,
        factory=partial(SingleValue, code=290),
        mindxf=DXF2000,
        maxdxf=DXF2018,
        priority=21000,
        default=0,
    ),
    # 0 = Disables SORTENTS
    # 1 = Sorts for object selection
    # 2 = Sorts for object snap
    # 4 = Sorts for redraws; obsolete
    # 8 = Sorts for MSLIDE command slide creation; obsolete
    # 16 = Sorts for REGEN commands
    # 32 = Sorts for plotting
    # 64 = Sorts for PostScript output; obsolete
    "$SORTENTS": HeaderVarDef(
        name="$SORTENTS",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21100,
        default=127,
    ),
    "$INDEXCTL": HeaderVarDef(
        name="$INDEXCTL",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21200,
        default=0,
    ),
    "$HIDETEXT": HeaderVarDef(
        name="$HIDETEXT",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21300,
        default=1,
    ),
    "$XCLIPFRAME": HeaderVarDef(
        name="$XCLIPFRAME",
        code=280,  # 2004 & 2007 = 290
        factory=partial(SingleValue, code=280),  # 2004 & 2007 = 290
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21400,
        # 0 = Clipping boundary is not visible
        # 1 = Clipping boundary is visible
        # default for 2004 & 2007 is 0; R2010+ is 2?
        default=1,  # changed 2014-03-16 issue #1049 - from 2 to 1
    ),
    "$HALOGAP": HeaderVarDef(
        name="$HALOGAP",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21500,
        default=0,
    ),
    "$OBSCOLOR": HeaderVarDef(
        name="$OBSCOLOR",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21600,
        default=257,
    ),
    "$OBSLTYPE": HeaderVarDef(
        name="$OBSLTYPE",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21700,
        default=0,
    ),
    "$INTERSECTIONDISPLAY": HeaderVarDef(
        name="$INTERSECTIONDISPLAY",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21800,
        default=0,
    ),
    "$INTERSECTIONCOLOR": HeaderVarDef(
        name="$INTERSECTIONCOLOR",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=21900,
        default=257,
    ),
    "$DIMASSOC": HeaderVarDef(
        name="$DIMASSOC",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=22000,
        default=2,
    ),
    "$PROJECTNAME": HeaderVarDef(
        name="$PROJECTNAME",
        code=1,
        factory=partial(SingleValue, code=1),
        mindxf=DXF2004,
        maxdxf=DXF2018,
        priority=22100,
        default="",
    ),
    "$CAMERADISPLAY": HeaderVarDef(
        name="$CAMERADISPLAY",
        code=290,
        factory=partial(SingleValue, code=290),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22200,
        default=0,
    ),
    "$LENSLENGTH": HeaderVarDef(
        name="$LENSLENGTH",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22300,
        default=50.0,
    ),
    "$CAMERAHEIGHT": HeaderVarDef(
        name="$CAMERAHEIGHT",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22400,
        default=0.0,
    ),
    "$STEPSPERSEC": HeaderVarDef(
        name="$STEPSPERSEC",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22500,
        default=24.0,
    ),
    "$STEPSIZE": HeaderVarDef(
        name="$STEPSIZE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22600,
        default=100.0,
    ),
    "$3DDWFPREC": HeaderVarDef(
        name="$3DDWFPREC",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22700,
        default=2.0,
    ),
    "$PSOLWIDTH": HeaderVarDef(
        name="$PSOLWIDTH",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22800,
        default=0.005,
    ),
    "$PSOLHEIGHT": HeaderVarDef(
        name="$PSOLHEIGHT",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=22900,
        default=0.08,
    ),
    "$LOFTANG1": HeaderVarDef(
        name="$LOFTANG1",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23000,
        default=1.570796326795,
    ),
    "$LOFTANG2": HeaderVarDef(
        name="$LOFTANG2",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23100,
        default=1.570796326795,
    ),
    "$LOFTMAG1": HeaderVarDef(
        name="$LOFTMAG1",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23200,
        default=0.0,
    ),
    "$LOFTMAG2": HeaderVarDef(
        name="$LOFTMAG2",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23300,
        default=0.0,
    ),
    "$LOFTPARAM": HeaderVarDef(
        name="$LOFTPARAM",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23400,
        default=7,
    ),
    "$LOFTNORMALS": HeaderVarDef(
        name="$LOFTNORMALS",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23500,
        default=1,
    ),
    "$LATITUDE": HeaderVarDef(
        name="$LATITUDE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23600,
        default=37.795,
    ),
    "$LONGITUDE": HeaderVarDef(
        name="$LONGITUDE",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23700,
        default=-122.394,
    ),
    "$NORTHDIRECTION": HeaderVarDef(
        name="$NORTHDIRECTION",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23800,
        default=0.0,
    ),
    "$TIMEZONE": HeaderVarDef(
        name="$TIMEZONE",
        code=70,
        factory=partial(SingleValue, code=70),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=23900,
        default=-8000,
    ),
    "$LIGHTGLYPHDISPLAY": HeaderVarDef(
        name="$LIGHTGLYPHDISPLAY",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24000,
        default=1,
    ),
    "$TILEMODELIGHTSYNCH": HeaderVarDef(
        name="$TILEMODELIGHTSYNCH",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24100,
        default=1,
    ),
    "$CMATERIAL": HeaderVarDef(
        name="$CMATERIAL",
        code=347,
        factory=partial(SingleValue, code=347),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24200,
        default="45",
    ),  # default value '0' crashes BricsCAD
    "$SOLIDHIST": HeaderVarDef(
        name="$SOLIDHIST",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24300,
        default=0,
    ),
    "$SHOWHIST": HeaderVarDef(
        name="$SHOWHIST",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24400,
        default=1,
    ),
    "$DWFFRAME": HeaderVarDef(
        name="$DWFFRAME",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24500,
        default=2,
    ),
    "$DGNFRAME": HeaderVarDef(
        name="$DGNFRAME",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24600,
        default=2,
    ),
    "$REALWORLDSCALE": HeaderVarDef(
        name="$REALWORLDSCALE",
        code=290,
        factory=partial(SingleValue, code=290),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24700,
        default=1,
    ),
    "$INTERFERECOLOR": HeaderVarDef(
        name="$INTERFERECOLOR",
        code=62,
        factory=partial(SingleValue, code=62),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24800,
        default=256,
    ),
    "$INTERFEREOBJVS": HeaderVarDef(
        # Handle to a VisualStyle, if pointing to an invalid or non-existing VS,
        # copy to clipboard in AutoCAD can fail.
        name="$INTERFEREOBJVS",
        code=345,
        factory=partial(SingleValue, code=345),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24810,
        default=None,
    ),  # will not be written, if not set
    "$INTERFEREVPVS": HeaderVarDef(
        # Handle to a VisualStyle, if pointing to an invalid or non-existing VS,
        # copy to clipboard in AutoCAD can fail.
        name="$INTERFEREVPVS",
        code=346,
        factory=partial(SingleValue, code=346),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24820,
        default=None,
    ),  # will not be written, if not set
    "$CSHADOW": HeaderVarDef(
        name="$CSHADOW",
        code=280,
        factory=partial(SingleValue, code=280),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=24900,
        default=0,
    ),
    "$SHADOWPLANELOCATION": HeaderVarDef(
        name="$SHADOWPLANELOCATION",
        code=40,
        factory=partial(SingleValue, code=40),
        mindxf=DXF2007,
        maxdxf=DXF2018,
        priority=25000,
        default=0.0,
    ),
}


def version_specific_group_code(name: str, dxfversion: str) -> int:
    group_code = HEADER_VAR_MAP[name].code
    # The HEADER_VAR_MAP contains the group codes for the latest DXF version.
    # This section adjust changed group codes for older DXF versions:
    if name == "$ACADMAINTVER":
        group_code = 70 if dxfversion < DXF2018 else 90
    elif name == "$XCLIPFRAME":
        group_code = 290 if dxfversion < DXF2010 else 280
    return group_code


VERSION_SPECIFIC_HEADER_VARS = """
DXF Version R2004
Name                 | BC GCode | BC Value | EZ GCode | EZ Value
----------------------------------------------------------------
$XCLIPFRAME          |      290 |        0 |      280 |        1

DXF Version R2007
Name                 | BC GCode | BC Value | EZ GCode | EZ Value
----------------------------------------------------------------
$XCLIPFRAME          |      290 |        0 |      280 |        1
"""



# Copyright (c) 2020-2021, Manfred Moitzi
# License: MIT License
from typing import cast, Union, List

import pytest
import ezdxf
import math

from ezdxf.entities import Ellipse, Point, Arc, DXFEntity, Insert
from ezdxf.math import Vec3


@pytest.fixture(scope="module")
def doc():
    d = ezdxf.new()
    blk = d.blocks.new("Test1")
    blk.add_line((0, 0), (1, 0))
    blk.add_line((0, 0), (0, 1))
    msp = d.modelspace()
    msp.add_blockref("Test1", (10, 10))
    msp.add_blockref(
        "Test1", (20, 10), dxfattribs={"xscale": 2}
    )  # yscale and zscale
    return d


@pytest.fixture(scope="module")
def msp(doc):
    return doc.modelspace()


@pytest.fixture(scope="module")
def entitydb(doc):
    return doc.entitydb


def test_virtual_entities_owner_and_handle(msp):
    blockrefs = msp.query("INSERT")
    blockref = blockrefs.first

    virtual_entities = list(blockref.virtual_entities())
    assert len(virtual_entities) == 2
    for e in virtual_entities:
        # Virtual entities should not be stored in the entity database.
        assert e.dxf.handle is None, "handle should be None"
        # Virtual entities should not reside in a layout.
        assert e.dxf.owner is None, "owner should be None"
        # Virtual entities should be assigned to the same document as the block reference.
        assert e.doc is blockref.doc


def test_virtual_entities_source_block_reference(msp):
    blockrefs = msp.query("INSERT")
    blockref = blockrefs.first

    for e in blockref.virtual_entities():
        entity = cast(DXFEntity, e)
        assert entity.has_source_block_reference is True
        assert entity.source_block_reference is blockref


def test_copying_entities_from_block_reference_removes_source_context(msp):
    """Copying an entity from a block reference, takes it out of context of
    this block reference!
    """
    blockrefs = msp.query("INSERT")
    for e in blockrefs.first.virtual_entities():
        copy = e.copy()
        assert copy.source_block_reference is None


def test_virtual_entities_transformation(msp):
    blockrefs = msp.query("INSERT")
    blockref = blockrefs[0]

    virtual_entities = list(blockref.virtual_entities())
    assert len(virtual_entities) == 2

    e = virtual_entities[0]
    assert e.dxftype() == "LINE"
    assert e.dxf.start == blockref.dxf.insert
    assert e.dxf.end == blockref.dxf.insert + (1, 0)

    e = virtual_entities[1]
    assert e.dxftype() == "LINE"
    assert e.dxf.start == blockref.dxf.insert
    assert e.dxf.end == blockref.dxf.insert + (0, 1)

    blockref = blockrefs[1]
    virtual_entities = list(blockref.virtual_entities())

    e = virtual_entities[0]
    assert e.dxftype() == "LINE"
    assert e.dxf.start == blockref.dxf.insert
    assert e.dxf.end == blockref.dxf.insert + (2, 0), "should apply xscale 2"

    e = virtual_entities[1]
    assert e.dxftype() == "LINE"
    assert e.dxf.start == blockref.dxf.insert
    assert e.dxf.end == blockref.dxf.insert + (0, 1), "should apply yscale 1"


def test_explode_blockrefs(doc, msp, entitydb):
    blockrefs = msp.query("INSERT")
    blockref = blockrefs.first
    blockref_owner = blockref.dxf.owner
    blockref_insert = blockref.dxf.insert

    assert len(msp) == 2  # 2 INSERT
    exploded_entities = blockref.explode()
    assert (
        blockref.is_alive is False
    ), "Exploded block reference should be destroyed."
    assert len(exploded_entities) == 2
    assert len(msp) == 3  # 2 INSERT - 1 exploded INSERT + 2 LINE

    e = exploded_entities[0]
    # Exploded entities should be stored in the entity database.
    assert e.dxf.handle is not None, "entity should have a handle"
    assert e.dxf.handle in entitydb
    # Exploded entities should reside in a layout.
    assert e.dxf.owner is not None, "entity should have an owner"
    assert e.dxf.owner is blockref_owner
    # Exploded entities should be assigned to the same document as the block reference.
    assert e.doc is doc

    assert e.dxftype() == "LINE"
    assert e.dxf.start == blockref_insert
    assert e.dxf.end == blockref_insert + (1, 0)

    e = exploded_entities[1]
    assert e.dxftype() == "LINE"
    assert e.dxf.start == blockref_insert
    assert e.dxf.end == blockref_insert + (0, 1)


def test_explode_polyline_bulge(doc, msp):
    blk = doc.blocks.new("Test03")
    blk.add_lwpolyline([(0, 0), (3, 0, 0.5), (6, 0), (9, 0)], format="xyb")
    block_ref = msp.add_blockref(
        "Test03",
        insert=(0, 0),
        dxfattribs={
            "yscale": 0.5,
        },
    )
    entities = list(block_ref.virtual_entities())
    assert len(entities) == 3

    e = entities[0]
    assert e.dxftype() == "LINE"
    assert e.dxf.start == (0, 0)
    assert e.dxf.end == (3, 0)

    e = entities[1]
    e = cast(Ellipse, e)
    assert e.dxftype() == "ELLIPSE"
    assert e.dxf.center.isclose((4.5, 0.5625, 0))
    assert e.dxf.major_axis.isclose((1.875, 0.0, 0))
    assert e.dxf.ratio == 0.5
    assert math.isclose(e.dxf.start_param, -2.498091544796509 % math.tau)
    assert math.isclose(e.dxf.end_param, -0.6435011087932843 % math.tau)
    assert e.start_point.isclose(Vec3(3, 0, 0))
    assert e.end_point.isclose(Vec3(6, 0, 0), abs_tol=1e-5)

    e = entities[2]
    assert e.dxftype() == "LINE"
    assert e.dxf.start == (6, 0)
    assert e.dxf.end == (9, 0)


def test_explode_blockref_with_attrib(msp):
    blockref = msp.add_blockref("Test1", (20, 10))  # with attrib
    blockref.add_attrib(tag="TAG", text="Text", insert=(1.5, 2.6))
    assert len(blockref.attribs) == 1, "Error in add_attrib()"
    attrib = blockref.attribs[0]

    exploded_entities = blockref.explode()
    assert (
        blockref.is_alive is False
    ), "Exploded block reference should be destroyed."
    assert attrib.is_alive is False, "Exploded attribs should be destroyed."
    assert len(exploded_entities) == 3, "2x LINE and 1x TEXT"
    text = exploded_entities[-1]
    assert text.dxftype() == "TEXT"
    assert text.dxf.text == "Text"
    assert text.dxf.insert == (1.5, 2.6), "ATTRIB already located in WCS"


def test_examine_uniform_scaled_ellipse(doc, msp):
    blk = doc.blocks.new("EllipseBlk")
    blk.add_ellipse((0, 0), major_axis=(2, 0), ratio=0.5)
    blkref = msp.add_blockref("EllipseBlk", insert=(2, 2)).set_scale(2)
    ellipse = list(blkref.virtual_entities())[0]
    assert ellipse.dxftype() == "ELLIPSE"
    assert ellipse.dxf.center == (2, 2)
    assert ellipse.dxf.major_axis == (4, 0)
    assert ellipse.dxf.ratio == 0.5


def test_skipped_entities_callback(doc, msp):
    blk = doc.blocks.new("test_block")
    hatch = blk.add_hatch()
    edge_path = hatch.paths.add_edge_path()
    edge_path.add_arc((0, 0))
    blk.add_line((0, 0), (1, 0))
    blkref = msp.add_blockref("test_block", insert=(0, 0)).place(
        (0, 0), (1, 2, 3)
    )
    skipped_entities = []

    def on_entity_skipped(entity, reason):
        skipped_entities.append((entity, reason))

    assert not blkref.has_uniform_scaling
    assert hatch.paths.has_critical_elements()
    entities = list(
        blkref.virtual_entities(skipped_entity_callback=on_entity_skipped)
    )

    assert len(entities) == 2
    assert entities[0].dxftype() == "HATCH"
    assert entities[1].dxftype() == "LINE"
    assert len(skipped_entities) == 0


def _get_transformed_curve(
    scale_factors: Vec3, rotation: float, is_arc: bool
) -> Union[Ellipse, Arc]:
    doc = ezdxf.new()
    blk = doc.blocks.new("block")
    if is_arc:
        blk.add_arc(
            (0, 0), radius=1, start_angle=0, end_angle=math.degrees(math.pi / 2)
        )
    else:
        blk.add_ellipse(
            (0, 0),
            major_axis=(1, 0),
            ratio=1,
            start_param=0,
            end_param=math.pi / 2,
        )

    assert blk[0].start_point.isclose(Vec3(1, 0, 0))
    assert blk[0].end_point.isclose(Vec3(0, 1, 0))

    blk.add_point((1, 0))
    blk.add_point((0, 1))
    block_ref = doc.modelspace().add_blockref(
        "block",
        insert=(0, 0),
        dxfattribs={
            "xscale": scale_factors.x,
            "yscale": scale_factors.y,
            "zscale": scale_factors.z,
            "rotation": math.degrees(rotation),
        },
    )
    entities = list(block_ref.virtual_entities())
    assert len(entities) == 3
    ellipse = cast(Union[Ellipse, Arc], entities[0])

    # points should have been transformed the same as the ellipse
    assert entities[1].dxftype() == "POINT"
    start_point = cast(Point, entities[1])
    assert start_point.dxf.location.isclose(ellipse.start_point)
    assert entities[2].dxftype() == "POINT"
    end_point = cast(Point, entities[2])
    assert end_point.dxf.location.isclose(ellipse.end_point)

    return ellipse


def _check_curve(
    ellipse: Ellipse,
    expected_start: Vec3,
    expected_end: Vec3,
    expected_extrusion: Vec3,
):
    assert ellipse.start_point.isclose(expected_start)
    assert ellipse.end_point.isclose(expected_end)
    assert ellipse.dxf.extrusion.is_parallel(expected_extrusion)


@pytest.mark.parametrize(
    "zscale,is_arc",
    [(1, False), (0.5, False), (1, True), (0.5, True), (-1, False), (-1, True)],
)
def test_rotated_and_reflected_curves(zscale, is_arc):
    scale = Vec3(1, 1, zscale)

    ellipse = _get_transformed_curve(scale, 0.0, is_arc)
    _check_curve(ellipse, Vec3(1, 0, 0), Vec3(0, 1, 0), Vec3(0, 0, zscale))

    ellipse = _get_transformed_curve(scale, math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, 1, 0), Vec3(-1, 0, 0), Vec3(0, 0, zscale))

    ellipse = _get_transformed_curve(scale, math.pi, is_arc)
    _check_curve(ellipse, Vec3(-1, 0, 0), Vec3(0, -1, 0), Vec3(0, 0, zscale))

    ellipse = _get_transformed_curve(scale, 3 * math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, -1, 0), Vec3(1, 0, 0), Vec3(0, 0, zscale))

    scale = Vec3(1, -1, zscale)

    ellipse = _get_transformed_curve(scale, 0.0, is_arc)
    _check_curve(ellipse, Vec3(1, 0, 0), Vec3(0, -1, 0), Vec3(0, 0, zscale))

    ellipse = _get_transformed_curve(scale, math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, 1, 0), Vec3(1, 0, 0), Vec3(0, 0, zscale))

    ellipse = _get_transformed_curve(scale, math.pi, is_arc)
    _check_curve(ellipse, Vec3(-1, 0, 0), Vec3(0, 1, 0), Vec3(0, 0, zscale))

    ellipse = _get_transformed_curve(scale, 3 * math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, -1, 0), Vec3(-1, 0, 0), Vec3(0, 0, zscale))

    scale = Vec3(-1, 1, zscale)

    ellipse = _get_transformed_curve(scale, 0.0, is_arc)
    _check_curve(ellipse, Vec3(-1, 0, 0), Vec3(0, 1, 0), Vec3(0, 0, zscale))

    ellipse = _get_transformed_curve(scale, math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, -1, 0), Vec3(-1, 0, 0), Vec3(0, 0, zscale))

    ellipse = _get_transformed_curve(scale, math.pi, is_arc)
    _check_curve(ellipse, Vec3(1, 0, 0), Vec3(0, -1, 0), Vec3(0, 0, zscale))

    ellipse = _get_transformed_curve(scale, 3 * math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, 1, 0), Vec3(1, 0, 0), Vec3(0, 0, zscale))

    scale = Vec3(-1, -1, zscale)

    ellipse = _get_transformed_curve(scale, 0.0, is_arc)
    _check_curve(ellipse, Vec3(-1, 0, 0), Vec3(0, -1, 0), Vec3(0, 0, zscale))

    ellipse = _get_transformed_curve(scale, math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, -1, 0), Vec3(1, 0, 0), Vec3(0, 0, zscale))

    ellipse = _get_transformed_curve(scale, math.pi, is_arc)
    _check_curve(ellipse, Vec3(1, 0, 0), Vec3(0, 1, 0), Vec3(0, 0, zscale))

    ellipse = _get_transformed_curve(scale, 3 * math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, 1, 0), Vec3(-1, 0, 0), Vec3(0, 0, zscale))


@pytest.mark.parametrize("stretch,is_arc", [(0.5, False), (0.5, True)])
def test_rotated_and_reflected_and_stretched_curves(stretch, is_arc):
    scale = Vec3(1, stretch, 1)

    ellipse = _get_transformed_curve(scale, 0.0, is_arc)
    _check_curve(ellipse, Vec3(1, 0, 0), Vec3(0, stretch, 0), Vec3(0, 0, 1))

    ellipse = _get_transformed_curve(scale, math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, 1, 0), Vec3(-stretch, 0, 0), Vec3(0, 0, 1))

    ellipse = _get_transformed_curve(scale, math.pi, is_arc)
    _check_curve(ellipse, Vec3(-1, 0, 0), Vec3(0, -stretch, 0), Vec3(0, 0, 1))

    ellipse = _get_transformed_curve(scale, 3 * math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, -1, 0), Vec3(stretch, 0, 0), Vec3(0, 0, 1))

    scale = Vec3(1, -stretch, 1)

    ellipse = _get_transformed_curve(scale, 0.0, is_arc)
    _check_curve(ellipse, Vec3(1, 0, 0), Vec3(0, -stretch, 0), Vec3(0, 0, 1))

    ellipse = _get_transformed_curve(scale, math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, 1, 0), Vec3(stretch, 0, 0), Vec3(0, 0, 1))

    ellipse = _get_transformed_curve(scale, math.pi, is_arc)
    _check_curve(ellipse, Vec3(-1, 0, 0), Vec3(0, stretch, 0), Vec3(0, 0, 1))

    ellipse = _get_transformed_curve(scale, 3 * math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, -1, 0), Vec3(-stretch, 0, 0), Vec3(0, 0, 1))

    scale = Vec3(-1, stretch, 1)

    ellipse = _get_transformed_curve(scale, 0.0, is_arc)
    _check_curve(ellipse, Vec3(-1, 0, 0), Vec3(0, stretch, 0), Vec3(0, 0, 1))

    ellipse = _get_transformed_curve(scale, math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, -1, 0), Vec3(-stretch, 0, 0), Vec3(0, 0, 1))

    ellipse = _get_transformed_curve(scale, math.pi, is_arc)
    _check_curve(ellipse, Vec3(1, 0, 0), Vec3(0, -stretch, 0), Vec3(0, 0, 1))

    ellipse = _get_transformed_curve(scale, 3 * math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, 1, 0), Vec3(stretch, 0, 0), Vec3(0, 0, 1))

    scale = Vec3(-1, -stretch, 1)

    ellipse = _get_transformed_curve(scale, 0.0, is_arc)
    _check_curve(ellipse, Vec3(-1, 0, 0), Vec3(0, -stretch, 0), Vec3(0, 0, 1))

    ellipse = _get_transformed_curve(scale, math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, -1, 0), Vec3(stretch, 0, 0), Vec3(0, 0, 1))

    ellipse = _get_transformed_curve(scale, math.pi, is_arc)
    _check_curve(ellipse, Vec3(1, 0, 0), Vec3(0, stretch, 0), Vec3(0, 0, 1))

    ellipse = _get_transformed_curve(scale, 3 * math.pi / 2, is_arc)
    _check_curve(ellipse, Vec3(0, 1, 0), Vec3(-stretch, 0, 0), Vec3(0, 0, 1))


def test_explode_xref(doc, msp):
    doc.add_xref_def("test.dxf", "xref_test")
    xref = msp.add_blockref("xref_test", (0, 0))
    assert xref.is_xref() is True

    result = xref.explode()
    assert len(result) == 0


class TestSourceBlockReferenceFromNestedBlockReferences:
    @pytest.fixture(scope="class")
    def doc(self):
        doc_ = ezdxf.new()
        blk0 = doc_.blocks.new("BLK0")
        blk1 = doc_.blocks.new("BLK1")
        blk2 = doc_.blocks.new("BLK2")

        blk2.add_point((2, 2))
        blk1.add_point((1, 1))
        # block reference 2
        blk1.add_blockref("BLK2", (2, 2))
        blk0.add_point((0, 0))
        # block reference 1
        blk0.add_blockref("BLK1", (1, 1))
        # block reference 0
        doc_.modelspace().add_blockref("BLK0", (0, 0))
        return doc_

    def test_virtual_entities_from_block_ref_0(self, doc):
        blkref0 = cast(Insert, doc.modelspace()[0])
        assert isinstance(blkref0, Insert) is True
        entities: List[DXFEntity] = list(blkref0.virtual_entities())
        point0 = cast(Point, entities[0])
        assert point0.dxftype() == "POINT"
        assert point0.source_block_reference is blkref0

        # An INSERT entity is a just a simple DXF entity
        blkref1 = cast(Insert, entities[1])
        assert blkref1.dxftype() == "INSERT"
        assert blkref1.source_block_reference is blkref0

    def test_virtual_entities_from_block_ref_1(self, doc):
        blkref0 = cast(Insert, doc.modelspace()[0])
        entities: List[DXFEntity] = list(blkref0.virtual_entities())
        blkref1 = cast(Insert, entities[1])
        entities = list(blkref1.virtual_entities())

        point1 = cast(Point, entities[0])
        assert point1.dxftype() == "POINT"
        assert point1.source_block_reference is blkref1

        blkref2 = cast(Insert, entities[1])
        assert blkref2.dxftype() == "INSERT"
        assert blkref2.source_block_reference is blkref1

    def test_virtual_entities_from_block_ref_2(self, doc):
        blkref0 = cast(Insert, doc.modelspace()[0])
        entities: List[DXFEntity] = list(blkref0.virtual_entities())
        blkref1 = cast(Insert, entities[1])
        entities = list(blkref1.virtual_entities())
        blkref2 = cast(Insert, entities[1])

        for point in blkref2.virtual_entities():
            assert point.dxftype() == "POINT"
            assert point.source_block_reference is blkref2


class TestEntitiesInRedrawOrder:
    @pytest.fixture(scope="class")
    def blk(self):
        d = ezdxf.new()
        blk = d.blocks.new("BLK0")
        line0 = blk.add_line((0, 0), (1, 0))
        line1 = blk.add_line((0, 0), (0, 1))
        line2 = blk.add_line((0, 0), (1, 1))
        # shuffle entities
        e = blk.entity_space.entities
        e[0] = line2
        e[1] = line0
        e[2] = line1
        msp = d.modelspace()
        msp.add_blockref("Test1", (10, 10))
        return blk

    def test_ascending_redraw_order(self, blk):
        entities = list(blk.entities_in_redraw_order())
        handles = [e.dxf.handle for e in entities]
        assert handles[0] < handles[1] < handles[2]

    def test_descending_redraw_order(self, blk):
        entities = list(blk.entities_in_redraw_order(reverse=True))
        handles = [e.dxf.handle for e in entities]
        assert handles[0] > handles[1] > handles[2]


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2024 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Sequence, Optional, Union, TYPE_CHECKING, Iterator
from typing_extensions import Self
import abc
import copy

from ezdxf.audit import Auditor, AuditError
from ezdxf.lldxf import const
from ezdxf.lldxf.tags import Tags
from ezdxf import colors
from ezdxf.tools import pattern
from ezdxf.math import Vec3, Matrix44
from ezdxf.math.transformtools import OCSTransform
from .boundary_paths import BoundaryPaths
from .dxfns import SubclassProcessor, DXFNamespace
from .dxfgfx import DXFGraphic
from .gradient import Gradient
from .pattern import Pattern, PatternLine
from .dxfentity import DXFEntity
from .copy import default_copy

if TYPE_CHECKING:
    from ezdxf import xref

RGB = colors.RGB

__all__ = ["DXFPolygon"]

PATH_CODES = {
    10,
    11,
    12,
    13,
    40,
    42,
    50,
    51,
    42,
    72,
    73,
    74,
    92,
    93,
    94,
    95,
    96,
    97,
    330,
}
PATTERN_DEFINITION_LINE_CODES = {53, 43, 44, 45, 46, 79, 49}


class DXFPolygon(DXFGraphic):
    """Base class for the HATCH and the MPOLYGON entity."""

    LOAD_GROUP_CODES: dict[int, Union[str, list[str]]] = {}

    def __init__(self) -> None:
        super().__init__()
        self.paths = BoundaryPaths()
        self.pattern: Optional[Pattern] = None
        self.gradient: Optional[Gradient] = None
        self.seeds: list[tuple[float, float]] = []  # not supported/exported by MPOLYGON

    def copy_data(self, entity: Self, copy_strategy=default_copy) -> None:
        """Copy paths, pattern, gradient, seeds."""
        assert isinstance(entity, DXFPolygon)
        entity.paths = copy.deepcopy(self.paths)
        entity.pattern = copy.deepcopy(self.pattern)
        entity.gradient = copy.deepcopy(self.gradient)
        entity.seeds = copy.deepcopy(self.seeds)

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            # Copy without subclass marker:
            tags = Tags(processor.subclasses[2][1:])
            # Removes boundary path data from tags:
            tags = self.load_paths(tags)
            # Removes gradient data from tags:
            tags = self.load_gradient(tags)
            # Removes pattern from tags:
            tags = self.load_pattern(tags)
            # Removes seeds from tags:
            tags = self.load_seeds(tags)

            # Load HATCH DXF attributes from remaining tags:
            processor.fast_load_dxfattribs(
                dxf, self.LOAD_GROUP_CODES, subclass=tags, recover=True
            )
        return dxf

    def load_paths(self, tags: Tags) -> Tags:
        # Find first group code 91 = count of loops, Spline data also contains
        # group code 91!
        try:
            start_index = tags.tag_index(91)
        except const.DXFValueError:
            raise const.DXFStructureError(
                f"{self.dxftype()}: Missing required DXF tag 'Number of "
                f"boundary paths (loops)' (code=91)."
            )

        path_tags = tags.collect_consecutive_tags(PATH_CODES, start=start_index + 1)
        if len(path_tags):
            self.paths = BoundaryPaths.load_tags(path_tags)
        end_index = start_index + len(path_tags) + 1
        del tags[start_index:end_index]
        return tags

    def load_pattern(self, tags: Tags) -> Tags:
        try:
            # Group code 78 = Number of pattern definition lines
            index = tags.tag_index(78)
        except const.DXFValueError:
            # No pattern definition lines found.
            return tags

        pattern_tags = tags.collect_consecutive_tags(
            PATTERN_DEFINITION_LINE_CODES, start=index + 1
        )
        self.pattern = Pattern.load_tags(pattern_tags)

        # Delete pattern data including length tag 78
        del tags[index : index + len(pattern_tags) + 1]
        return tags

    def load_gradient(self, tags: Tags) -> Tags:
        try:
            index = tags.tag_index(450)
        except const.DXFValueError:
            # No gradient data present
            return tags

        # Gradient data is always at the end of the AcDbHatch subclass.
        self.gradient = Gradient.load_tags(tags[index:])  # type: ignore
        # Remove gradient data from tags
        del tags[index:]
        return tags

    def map_resources(self, clone: Self, mapping: xref.ResourceMapper) -> None:
        """Translate resources from self to the copied entity."""
        assert isinstance(clone, DXFPolygon)
        assert clone.doc is not None

        super().map_resources(clone, mapping)
        db = clone.doc.entitydb
        for path in clone.paths:
            handles = [mapping.get_handle(h) for h in path.source_boundary_objects]
            path.source_boundary_objects = [h for h in handles if h in db]

    def load_seeds(self, tags: Tags) -> Tags:
        return tags

    @property
    def has_solid_fill(self) -> bool:
        """``True`` if entity has a solid fill. (read only)"""
        return bool(self.dxf.solid_fill)

    @property
    def has_pattern_fill(self) -> bool:
        """``True`` if entity has a pattern fill. (read only)"""
        return not bool(self.dxf.solid_fill)

    @property
    def has_gradient_data(self) -> bool:
        """``True`` if entity has a gradient fill. A hatch with gradient fill
        has also a solid fill. (read only)
        """
        return bool(self.gradient)

    @property
    def bgcolor(self) -> Optional[RGB]:
        """
        Set pattern fill background color as (r, g, b)-tuple, rgb values
        in the range [0, 255] (read/write/del)

        usage::

            r, g, b = entity.bgcolor  # get pattern fill background color
            entity.bgcolor = (10, 20, 30)  # set pattern fill background color
            del entity.bgcolor  # delete pattern fill background color

        """
        try:
            xdata_bgcolor = self.get_xdata("HATCHBACKGROUNDCOLOR")
        except const.DXFValueError:
            return None
        color = xdata_bgcolor.get_first_value(1071, 0)
        try:
            return colors.int2rgb(int(color))
        except ValueError:  # invalid data type
            return RGB(0, 0, 0)

    @bgcolor.setter
    def bgcolor(self, rgb: RGB) -> None:
        color_value = (
            colors.rgb2int(rgb) | -0b111110000000000000000000000000
        )  # it's magic

        self.discard_xdata("HATCHBACKGROUNDCOLOR")
        self.set_xdata("HATCHBACKGROUNDCOLOR", [(1071, color_value)])

    @bgcolor.deleter
    def bgcolor(self) -> None:
        self.discard_xdata("HATCHBACKGROUNDCOLOR")

    def set_gradient(
        self,
        color1: RGB = RGB(0, 0, 0),
        color2: RGB = RGB(255, 255, 255),
        rotation: float = 0.0,
        centered: float = 0.0,
        one_color: int = 0,
        tint: float = 0.0,
        name: str = "LINEAR",
    ) -> None:
        """Sets the gradient fill mode and removes all pattern fill related data, requires
        DXF R2004 or newer.  A gradient filled hatch is also a solid filled hatch.

        Valid gradient type names are:

            - "LINEAR"
            - "CYLINDER"
            - "INVCYLINDER"
            - "SPHERICAL"
            - "INVSPHERICAL"
            - "HEMISPHERICAL"
            - "INVHEMISPHERICAL"
            - "CURVED"
            - "INVCURVED"

        Args:
            color1: (r, g, b)-tuple for first color, rgb values as int in
                the range [0, 255]
            color2: (r, g, b)-tuple for second color, rgb values as int in
                the range [0, 255]
            rotation: rotation angle in degrees
            centered: determines whether the gradient is centered or not
            one_color: 1 for gradient from `color1` to tinted `color1`
            tint: determines the tinted target `color1` for a one color
                gradient. (valid range 0.0 to 1.0)
            name: name of gradient type, default "LINEAR"

        """
        if self.doc is not None and self.doc.dxfversion < const.DXF2004:
            raise const.DXFVersionError("Gradient support requires DXF R2004")
        if name and name not in const.GRADIENT_TYPES:
            raise const.DXFValueError(f"Invalid gradient type name: {name}")

        self.pattern = None
        self.dxf.solid_fill = 1
        self.dxf.pattern_name = "SOLID"
        self.dxf.pattern_type = const.HATCH_TYPE_PREDEFINED

        gradient = Gradient()
        gradient.color1 = color1
        gradient.color2 = color2
        gradient.one_color = one_color
        gradient.rotation = rotation
        gradient.centered = centered
        gradient.tint = tint
        gradient.name = name
        self.gradient = gradient

    def set_pattern_fill(
        self,
        name: str,
        color: int = 7,
        angle: float = 0.0,
        scale: float = 1.0,
        double: int = 0,
        style: int = 1,
        pattern_type: int = 1,
        definition=None,
    ) -> None:
        """Sets the pattern fill mode and removes all gradient related data.

        The pattern definition should be designed for a scale factor 1 and a rotation
        angle of 0 degrees.  The predefined hatch pattern like "ANSI33" are scaled
        according to the HEADER variable $MEASUREMENT for ISO measurement (m, cm, ... ),
        or imperial units (in, ft, ...), this replicates the behavior of BricsCAD.

        Args:
            name: pattern name as string
            color: pattern color as :ref:`ACI`
            angle: pattern rotation angle in degrees
            scale: pattern scale factor
            double: double size flag
            style: hatch style (0 = normal; 1 = outer; 2 = ignore)
            pattern_type: pattern type (0 = user-defined;
                1 = predefined; 2 = custom)
            definition: list of definition lines and a definition line is a
                4-tuple [angle, base_point, offset, dash_length_items],
                see :meth:`set_pattern_definition`

        """
        self.gradient = None
        self.dxf.solid_fill = 0
        self.dxf.pattern_name = name
        self.dxf.color = color
        self.dxf.pattern_scale = float(scale)
        self.dxf.pattern_angle = float(angle)
        self.dxf.pattern_double = int(double)
        self.dxf.hatch_style = style
        self.dxf.pattern_type = pattern_type

        if definition is None:
            measurement = 1
            if self.doc:
                measurement = self.doc.header.get("$MEASUREMENT", measurement)
            predefined_pattern = (
                pattern.ISO_PATTERN if measurement else pattern.IMPERIAL_PATTERN
            )
            definition = predefined_pattern.get(name, predefined_pattern["ANSI31"])
        self.set_pattern_definition(
            definition,
            factor=self.dxf.pattern_scale,
            angle=self.dxf.pattern_angle,
        )

    def set_pattern_definition(
        self, lines: Sequence, factor: float = 1, angle: float = 0
    ) -> None:
        """Setup pattern definition by a list of definition lines and the
        definition line is a 4-tuple (angle, base_point, offset, dash_length_items).
        The pattern definition should be designed for a pattern scale factor of 1 and
        a pattern rotation angle of 0.

            - angle: line angle in degrees
            - base-point: (x, y) tuple
            - offset: (dx, dy) tuple
            - dash_length_items: list of dash items (item > 0 is a line,
              item < 0 is a gap and item == 0.0 is a point)

        Args:
            lines: list of definition lines
            factor: pattern scale factor
            angle: rotation angle in degrees

        """
        if factor != 1 or angle:
            lines = pattern.scale_pattern(lines, factor=factor, angle=angle)
        self.pattern = Pattern(
            [PatternLine(line[0], line[1], line[2], line[3]) for line in lines]
        )

    def set_pattern_scale(self, scale: float) -> None:
        """Sets the pattern scale factor and scales the pattern definition.

        The method always starts from the original base scale, the
        :code:`set_pattern_scale(1)` call resets the pattern scale to the original
        appearance as defined by the pattern designer, but only if the pattern attribute
        :attr:`dxf.pattern_scale` represents the actual scale, it cannot
        restore the original pattern scale from the pattern definition itself.

        Args:
            scale: pattern scale factor

        """
        if not self.has_pattern_fill:
            return
        dxf = self.dxf
        self.pattern.scale(factor=1.0 / dxf.pattern_scale * scale)  # type: ignore
        dxf.pattern_scale = scale

    def set_pattern_angle(self, angle: float) -> None:
        """Sets the pattern rotation angle and rotates the pattern definition.

        The method always starts from the original base rotation of 0, the
        :code:`set_pattern_angle(0)` call resets the pattern rotation angle to the
        original appearance as defined by the pattern designer, but only if the
        pattern attribute :attr:`dxf.pattern_angle` represents the actual pattern
        rotation, it cannot restore the original rotation angle from the
        pattern definition itself.

        Args:
            angle: pattern rotation angle in degrees

        """
        if not self.has_pattern_fill:
            return
        dxf = self.dxf
        self.pattern.scale(angle=angle - dxf.pattern_angle)  # type: ignore
        dxf.pattern_angle = angle % 360.0

    def transform(self, m: Matrix44) -> DXFPolygon:
        """Transform entity by transformation matrix `m` inplace."""
        dxf = self.dxf
        ocs = OCSTransform(dxf.extrusion, m)

        elevation = Vec3(dxf.elevation).z
        self.paths.transform(ocs, elevation=elevation)
        dxf.elevation = ocs.transform_vertex(Vec3(0, 0, elevation)).replace(
            x=0.0, y=0.0
        )
        dxf.extrusion = ocs.new_extrusion
        if self.pattern:
            # todo: non-uniform scaling
            # take the scaling factor of the x-axis
            factor = ocs.transform_length((self.dxf.pattern_scale, 0, 0))
            angle = ocs.transform_deg_angle(self.dxf.pattern_angle)
            # todo: non-uniform pattern scaling is not supported
            self.pattern.scale(factor, angle)
            self.dxf.pattern_scale = factor
            self.dxf.pattern_angle = angle
        self.post_transform(m)
        return self

    def triangulate(self, max_sagitta, min_segments=16) -> Iterator[Sequence[Vec3]]:
        """Triangulate the HATCH/MPOLYGON in OCS coordinates, Elevation and offset is
        applied to all vertices.

        Args:
            max_sagitta: maximum distance from the center of the curve to the
                center of the line segment between two approximation points to determine
                if a segment should be subdivided.
            min_segments: minimum segment count per Bézier curve

        .. versionadded:: 1.1

        """
        from ezdxf import path

        elevation = Vec3(self.dxf.elevation)
        if self.dxf.hasattr("offset"):  # MPOLYGON
            elevation += Vec3(self.dxf.offset)  # offset in OCS?
        boundary_paths = [path.from_hatch_boundary_path(p) for p in self.paths]
        for vertices in path.triangulate(boundary_paths, max_sagitta, min_segments):
            yield tuple(elevation + v for v in vertices)

    def render_pattern_lines(self) -> Iterator[tuple[Vec3, Vec3]]:
        """Yields the pattern lines in WCS coordinates.

        .. versionadded:: 1.1

        """
        from ezdxf.render import hatching

        if self.has_pattern_fill:
            try:
                yield from hatching.hatch_entity(self)
            except hatching.HatchingError:
                return

    @abc.abstractmethod
    def set_solid_fill(self, color: int = 7, style: int = 1, rgb: Optional[RGB] = None):
        ...
    
    def audit(self, auditor: Auditor) -> None:
        super().audit(auditor)
        if not self.is_alive:
            return
        if not self.paths.is_valid():
            auditor.fixed_error(
                code=AuditError.INVALID_HATCH_BOUNDARY_PATH,
                message=f"Deleted entity {str(self)} containing invalid boundary paths."
            )
            auditor.trash(self)


# Copyright (c) 2014-2022, Manfred Moitzi
# License: MIT License
"""
DXF Types
=========

Required DXF tag interface:

    - property :attr:`code`: group code as int
    - property :attr:`value`: tag value of unspecific type
    - :meth:`dxfstr`: returns the DXF string
    - :meth:`clone`: returns a deep copy of tag

"""
from __future__ import annotations
from typing import (
    Union,
    Iterable,
    Sequence,
    Type,
    Any,
)
from array import array
from itertools import chain
from binascii import unhexlify, hexlify
import reprlib
from ezdxf.math import Vec3


TAG_STRING_FORMAT = "%3d\n%s\n"
POINT_CODES = {
    10,
    11,
    12,
    13,
    14,
    15,
    16,
    17,
    18,
    110,
    111,
    112,
    210,
    211,
    212,
    213,
    1010,
    1011,
    1012,
    1013,
}

MAX_GROUP_CODE = 1071
GENERAL_MARKER = 0
SUBCLASS_MARKER = 100
XDATA_MARKER = 1001
EMBEDDED_OBJ_MARKER = 101
APP_DATA_MARKER = 102
EXT_DATA_MARKER = 1001
GROUP_MARKERS = {
    GENERAL_MARKER,
    SUBCLASS_MARKER,
    EMBEDDED_OBJ_MARKER,
    APP_DATA_MARKER,
    EXT_DATA_MARKER,
}
BINARY_FLAGS = {70, 90}
HANDLE_CODES = {5, 105}
POINTER_CODES = set(chain(range(320, 370), range(390, 400), (480, 481, 1005)))

# pointer group codes 320-329 are not translated during INSERT and XREF operations
TRANSLATABLE_POINTER_CODES = set(
    chain(range(330, 370), range(390, 400), (480, 481, 1005))
)
HEX_HANDLE_CODES = set(chain(HANDLE_CODES, POINTER_CODES))
BINARY_DATA = {310, 311, 312, 313, 314, 315, 316, 317, 318, 319, 1004}
EMBEDDED_OBJ_STR = "Embedded Object"

BYTES = set(range(290, 300))  # bool

INT16 = set(
    chain(
        range(60, 80),
        range(170, 180),
        range(270, 290),
        range(370, 390),
        range(400, 410),
        range(1060, 1071),
    )
)

INT32 = set(
    chain(
        range(90, 100),
        range(420, 430),
        range(440, 450),
        range(450, 460),  # Long in DXF reference, ->signed<- or unsigned?
        [1071],
    )
)

INT64 = set(range(160, 170))

DOUBLE = set(
    chain(
        range(10, 60),
        range(110, 150),
        range(210, 240),
        range(460, 470),
        range(1010, 1060),
    )
)

VALID_XDATA_GROUP_CODES = {
    1000,
    1001,
    1002,
    1003,
    1004,
    1005,
    1010,
    1011,
    1012,
    1013,
    1040,
    1041,
    1042,
    1070,
    1071,
}


def _build_type_table(types):
    table = {}
    for caster, codes in types:
        for code in codes:
            table[code] = caster
    return table


TYPE_TABLE = _build_type_table(
    [
        # all group code < 0 are spacial tags for internal use
        (float, DOUBLE),
        (int, BYTES),
        (int, INT16),
        (int, INT32),
        (int, INT64),
    ]
)


class DXFTag:
    """Immutable DXFTag class.

    Args:
        code: group code as int
        value: tag value, type depends on group code

    """

    __slots__ = ("_code", "_value")

    def __init__(self, code: int, value: Any):
        self._code: int = int(code)
        # Do not use _value, always use property value - overwritten in subclasses
        self._value = value

    def __str__(self) -> str:
        """Returns content string ``'(code, value)'``."""
        return str((self._code, self.value))

    def __repr__(self) -> str:
        """Returns representation string ``'DXFTag(code, value)'``."""
        return f"DXFTag{str(self)}"

    @property
    def code(self) -> int:
        return self._code

    @property
    def value(self) -> Any:
        return self._value

    def __getitem__(self, index: int):
        """Returns :attr:`code` for index 0 and :attr:`value` for index 1,
        emulates a tuple.
        """
        return (self._code, self.value)[index]

    def __iter__(self):
        """Returns (code, value) tuples."""
        yield self._code
        yield self.value

    def __eq__(self, other) -> bool:
        """``True`` if `other` and `self` has same content for :attr:`code`
        and :attr:`value`.
        """
        return (self._code, self.value) == other

    def __hash__(self):
        """Hash support, :class:`DXFTag` can be used in sets and as dict key."""
        return hash((self._code, self._value))

    def dxfstr(self) -> str:
        """Returns the DXF string e.g. ``'  0\\nLINE\\n'``"""
        return TAG_STRING_FORMAT % (self.code, self._value)

    def clone(self) -> "DXFTag":
        """Returns a clone of itself, this method is necessary for the more
        complex (and not immutable) DXF tag types.
        """
        return self  # immutable tags


# Special marker tag
NONE_TAG = DXFTag(0, 0)


def uniform_appid(appid: str) -> str:
    if appid[0] == "{":
        return appid
    else:
        return "{" + appid


def is_app_data_marker(tag: DXFTag) -> bool:
    return tag.code == APP_DATA_MARKER and tag.value.startswith("{")


def is_embedded_object_marker(tag: DXFTag) -> bool:
    return tag.code == EMBEDDED_OBJ_MARKER and tag.value == EMBEDDED_OBJ_STR


def is_arbitrary_pointer(tag: DXFTag) -> bool:
    """Arbitrary object handles; handle values that are taken "as is".
    They are not translated during INSERT and XREF operations.
    """
    return 319 < tag.code < 330


def is_soft_pointer(tag: DXFTag) -> bool:
    """Soft-pointer handle; arbitrary soft pointers to other objects within same DXF
    file or drawing. Translated during INSERT and XREF operations.
    """
    return 329 < tag.code < 340 or tag.code == 1005


def is_hard_pointer(tag: DXFTag) -> bool:
    """Hard-pointer handle; arbitrary hard pointers to other objects within same DXF
    file or drawing. Translated during INSERT and XREF operations. Hard pointers
    protect an object from being purged.
    """
    code = tag.code
    return 339 < code < 350 or 389 < code < 400 or 479 < code < 482


def is_soft_owner(tag: DXFTag) -> bool:
    """Soft-owner handle; arbitrary soft ownership links to other objects within same
    DXF file or drawing. Translated during INSERT and XREF operations.
    """
    return 349 < tag.code < 360


def is_hard_owner(tag: DXFTag) -> bool:
    """Hard-owner handle; arbitrary hard ownership links to other objects within same
    DXF file or drawing. Translated during INSERT and XREF operations. Hard owner handle
    protect an object from being purged.
    """
    return 359 < tag.code < 370


def is_translatable_pointer(tag: DXFTag) -> bool:
    # pointer group codes 320-329 are not translated during INSERT and XREF operations
    return tag.code in TRANSLATABLE_POINTER_CODES


class DXFVertex(DXFTag):
    """Represents a 2D or 3D vertex, stores only the group code of the
    x-component of the vertex, because the y-group-code is x-group-code + 10
    and z-group-code id x-group-code+20, this is a rule that ALWAYS applies.
    This tag is `immutable` by design, not by implementation.

    Args:
        code: group code of x-component
        value: sequence of x, y and optional z values

    """

    __slots__ = ()

    def __init__(self, code: int, value: Iterable[float]):
        super(DXFVertex, self).__init__(code, array("d", value))

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return f"DXFVertex({self.code}, {str(self)})"

    def __hash__(self):
        x, y, *z = self._value
        z = 0.0 if len(z) == 0 else z[0]
        return hash((self.code, x, y, z))

    @property
    def value(self) -> tuple[float, ...]:
        return tuple(self._value)

    def dxftags(self) -> Iterable[DXFTag]:
        """Returns all vertex components as single :class:`DXFTag` objects."""
        c = self.code
        return (
            DXFTag(code, value) for code, value in zip((c, c + 10, c + 20), self.value)
        )

    def dxfstr(self) -> str:
        """Returns the DXF string for all vertex components."""
        return "".join(tag.dxfstr() for tag in self.dxftags())


class DXFBinaryTag(DXFTag):
    """Immutable BinaryTags class - immutable by design, not by implementation."""

    __slots__ = ()

    def __str__(self) -> str:
        return f"({self.code}, {self.tostring()})"

    def __repr__(self) -> str:
        return f"DXFBinaryTag({self.code}, {reprlib.repr(self.tostring())})"

    def tostring(self) -> str:
        """Returns binary value as single hex-string."""
        assert isinstance(self.value, bytes)
        return hexlify(self.value).upper().decode()

    def dxfstr(self) -> str:
        """Returns the DXF string for all vertex components."""
        return TAG_STRING_FORMAT % (self.code, self.tostring())

    @classmethod
    def from_string(cls, code: int, value: Union[str, bytes]):
        return cls(code, unhexlify(value))


def dxftag(code: int, value: Any) -> DXFTag:
    """DXF tag factory function.

    Args:
        code: group code
        value: tag value

    Returns: :class:`DXFTag` or inherited

    """
    if code in BINARY_DATA:
        return DXFBinaryTag(code, value)
    elif code in POINT_CODES:
        return DXFVertex(code, value)
    else:
        return DXFTag(code, cast_tag_value(code, value))


def tuples_to_tags(iterable: Iterable[tuple[int, Any]]) -> Iterable[DXFTag]:
    """Returns an iterable if :class:`DXFTag` or inherited, accepts an
    iterable of (code, value) tuples as input.
    """
    for code, value in iterable:
        if code in POINT_CODES:
            yield DXFVertex(code, value)
        elif code in BINARY_DATA:
            assert isinstance(value, (str, bytes))
            yield DXFBinaryTag.from_string(code, value)
        else:
            yield DXFTag(code, value)


def is_valid_handle(handle) -> bool:
    if isinstance(handle, str):
        try:
            int(handle, 16)
            return True
        except (ValueError, TypeError):
            pass
    return False


def is_binary_data(code: int) -> bool:
    return code in BINARY_DATA


def is_pointer_code(code: int) -> bool:
    return code in POINTER_CODES


def is_point_code(code: int) -> bool:
    return code in POINT_CODES


def is_point_tag(tag: Sequence) -> bool:
    return tag[0] in POINT_CODES


def cast_tag_value(code: int, value: Any) -> Any:
    return TYPE_TABLE.get(code, str)(value)


def tag_type(code: int) -> Type:
    return TYPE_TABLE.get(code, str)


def strtag(tag: Union[DXFTag, tuple[int, Any]]) -> str:
    return TAG_STRING_FORMAT % tuple(tag)


def get_xcode_for(code) -> int:
    if code in HEX_HANDLE_CODES:
        return 1005
    if code in BINARY_DATA:
        return 1004
    type_ = TYPE_TABLE.get(code, str)
    if type_ is int:
        return 1070
    if type_ is float:
        return 1040
    return 1000


def cast_value(code: int, value):
    if value is not None:
        if code in POINT_CODES:
            return Vec3(value)
        return TYPE_TABLE.get(code, str)(value)
    else:
        return None


TAG_TYPES = {
    int: "<int>",
    float: "<float>",
    str: "<str>",
}


def tag_type_str(code: int) -> str:
    if code in GROUP_MARKERS:
        return "<ctrl>"
    elif code in HANDLE_CODES:
        return "<handle>"
    elif code in POINTER_CODES:
        return "<ref>"
    elif is_point_code(code):
        return "<point>"
    elif is_binary_data(code):
        return "<bin>"
    else:
        return TAG_TYPES[tag_type(code)]


def render_tag(tag: DXFTag, col: int) -> Any:
    code, value = tag
    if col == 0:
        return str(code)
    elif col == 1:
        return tag_type_str(code)
    elif col == 2:
        return str(value)
    else:
        raise IndexError(col)



# Copyright (c) 2013-2026, Manfred Moitzi
# License: MIT-License
# mypy: ignore_errors=True
"""
EntityQueryParser implemented with the pyparsing module created by Paul T. McGuire.

QueryString := EntityQuery ("[" AttribQuery "]" "i"?)*

The QueryString consist of two queries, first the required entity query and the second optional attribute query,
enclosed in square brackets an optional appended "i" indicates ignore case for string queries.

1. EntityQuery (required)

The EntityQuery is a whitespace separated list of names (DXF entity names) or the special name "*".

2. AttribQuery (optional)

The AttribQuery is a boolean expression, supported boolean operators are:

  - '!' not: !term if true, if tern is false.
  - '&' and: term & term is true, if both terms are true.
  - '|' or: term | term is true, if one term is true.

The query itself consist of a name a comparator and a value, like "color < 7".
Supported comparators are:

  - '==': equal
  - '!=': not equal
  - '<': lower than
  - '<=': lower or equal than
  - '>': greater than
  - '>=': greater or equal than
  - '?': match a regular expression
  - '!?': does not match a regular expression

Values can be integers, floats or strings, strings have to be quoted ("I am a string" or 'I am a string').

examples:
    'LINE CIRCLE[layer=="construction"]' => all LINE and CIRCLE entities on layer "construction"
    '*[!(layer=="construction" & color<7)]' => all entities except those on layer == "construction" and color < 7
    'LINE[layer=="construction"]i' => all lines on layer named 'construction' with case insensitivity, "CoNsTrUcTiOn" is valid
"""

from pyparsing import *

LBRK = Suppress("[")
RBRK = Suppress("]")

number = Regex(r"[+-]?\d+(:?\.\d*)?(:?[eE][+-]?\d+)?")
number.add_parse_action(lambda t: float(t[0]))  # convert to float
string_ = quotedString.add_parse_action(lambda t: t[0][1:-1])  # remove quotes

EntityName = Word(alphanums + "_")
# ExcludeEntityName = Word(alphanums + '_!')
ExcludeEntityName = Regex(r"[!][\w]+")
AttribName = Word(alphanums + "_")
Relation = one_of(["==", "!=", "<", "<=", ">", ">=", "?", "!?"])

AttribValue = string_ | number
AttribQuery = Group(AttribName + Relation + AttribValue)
EntityNames = Group(
    (Literal("*") + ZeroOrMore(ExcludeEntityName)) | OneOrMore(EntityName)
).set_results_name("EntityQuery")

InfixBoolQuery = infix_notation(
    AttribQuery,
    [
        ("!", 1, opAssoc.RIGHT),
        ("&", 2, opAssoc.LEFT),
        ("|", 2, opAssoc.LEFT),
    ],
).set_results_name("AttribQuery")

AttribQueryOptions = Literal("i").set_results_name("AttribQueryOptions")

EntityQueryParser = EntityNames + Optional(
    LBRK + InfixBoolQuery + RBRK + Optional(AttribQueryOptions)
)



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
import pytest

import ezdxf
from ezdxf.entities import MPolygon
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text


# MPolygon is similar to Hatch
@pytest.fixture
def entity():
    return MPolygon.from_text(MPOLYGON_NO_FILL)


def test_if_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "MPOLYGON" in ENTITY_CLASSES


def test_default_init():
    entity = MPolygon()
    assert entity.dxftype() == "MPOLYGON"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = MPolygon.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": 7,  # color of boundary paths!
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.version == 1
    assert entity.dxf.solid_fill == 0
    assert entity.dxf.fill_color == ezdxf.const.BYLAYER


def test_fill_properties_without_solid_filling():
    entity = MPolygon()
    entity.dxf.solid_fill = 0
    assert entity.has_solid_fill is False
    assert entity.has_pattern_fill is True


def test_fill_properties_with_solid_filling():
    entity = MPolygon()
    entity.dxf.solid_fill = 1
    assert entity.has_solid_fill is True
    assert entity.has_pattern_fill is False


def test_load_from_text(entity):
    assert entity.dxf.layer == "mpolygons"
    assert entity.dxf.color == 1
    assert entity.dxf.version == 1
    assert entity.dxf.solid_fill == 0
    assert len(entity.paths) == 2
    assert entity.dxf.pattern_name == ""


def test_write_dxf_no_fill(entity):
    entity = MPolygon.from_text(MPOLYGON_NO_FILL)
    result = TagCollector.dxftags(entity, dxfversion=ezdxf.const.DXF2000)
    expected = basic_tags_from_text(MPOLYGON_NO_FILL)
    assert result == expected
    assert result.get_first_value(71) == 0  # pattern fill
    tags = list(result.pop_tags([52, 41, 77, 78]))
    assert tags == [
        (52, 0),  # pattern_angle tag must be presents
        (41, 1),  # pattern_scale tag must be presents
        (77, 0),  # pattern_double tag must be presents
        (78, 0),  # pattern length tag must be presents
    ], "required pattern tags are not in expected order"

    assert (
        result.has_tag(450) is False
    ), "gradient data is not supported for DXF R2000"


def test_write_dxf_r2004_no_fill_requires_basic_gradient_data(entity):
    result = TagCollector.dxftags(entity, dxfversion=ezdxf.const.DXF2004)
    tags = list(result.pop_tags([450, 451, 460, 461, 452, 462, 453, 470]))
    assert tags == [
        (450, 0),  # kind = solid fill
        (451, 0),  # reserved for the future
        (460, 0),  # angle in radians
        (461, 0),  # centered
        (452, 0),  # one color
        (462, 0),  # tint
        (453, 0),  # number of colors
        (470, ""),  # gradient name
    ], "required gradient tags are not in expected order"


def test_write_dxf_with_fill(entity):
    entity.dxf.solid_fill = 1
    entity.dxf.fill_color = 163
    result = TagCollector.dxftags(entity, dxfversion=ezdxf.const.DXF2000)
    assert result.get_first_value(71) == 1  # solid_fill
    assert (
        result.has_tag(52) is False
    ), "pattern_angle tag should not be presents"
    assert (
        result.has_tag(41) is False
    ), "pattern_scale tag should not be presents"
    assert (
        result.has_tag(77) is False
    ), "pattern_double tag should not be presents"
    assert (
        result.has_tag(78) is False
    ), "pattern length tag should not be presents"
    assert (
        result.has_tag(63) is False
    ), "fill color tag is not supported for DXF R2000"


def test_write_dxf_R2004_with_fill(entity):
    entity.dxf.solid_fill = 1
    entity.dxf.fill_color = 163
    result = TagCollector.dxftags(entity, dxfversion=ezdxf.const.DXF2004)
    assert result.get_first_value(63) == 163, "missing fill color tag"

    tags = list(result.pop_tags([450, 451, 460, 461, 452, 462, 453, 470]))
    assert tags == [
        (450, 0),  # kind = solid fill
        (451, 0),  # reserved for the future
        (460, 0),  # angle in radians
        (461, 0),  # centered
        (452, 0),  # one color
        (462, 0),  # tint
        (453, 0),  # number of colors
        (470, ""),  # gradient name
    ], "required gradient tags are not in expected order"


def test_write_correct_polyline_path_tag_order(entity):
    result = TagCollector.dxftags(entity)
    tags = list(result.pop_tags([92, 72, 73]))
    # 92 = path type 2: polyline path
    # 73 = is_closed
    # 72 = has_bulge
    # The group codes 73 and 72 are swapped in comparison to HATCH
    # contains 2 polyline paths
    assert tags == [(92, 2), (73, 0), (72, 0), (92, 2), (73, 0), (72, 0)]


def test_write_dxf_with_pattern_fill(entity):
    entity.set_pattern_fill("ANSI33", color=7, scale=0.01)
    result = TagCollector.dxftags(entity, dxfversion=ezdxf.const.DXF2000)
    assert result.has_tag(75) is False, "hatch style tag not supported?!"


def test_do_not_export_mpolygon_with_edge_paths(entity):
    # Edge paths are not supported by MPOLYGON as far as I known!
    entity.paths.add_edge_path()
    assert entity.preprocess_export(TagCollector()) is False


MPOLYGON_NO_FILL = """0
MPOLYGON
5
ABBA
330
DEAD
100
AcDbEntity
8
mpolygons
62
1
100
AcDbMPolygon
70
1
10
0.0
20
0.0
30
0.0
210
0.0
220
0.0
230
1.0
2

71
0
91
2
92
2
73
0
72
0
93
5
10
-85.2
20
35.04
10
-85.2
20
35.04
10
-85.2
20
35.04
10
-85.2
20
35.04
10
-85.2
20
35.04
92
2
73
0
72
0
93
4
10
-85.2
20
35.04
10
-85.2
20
35.04
10
-85.2
20
35.04
10
-85.2
20
35.04
76
0
52
0.0
41
1.0
77
0
78
0
11
0.0
21
0.0
99
0
"""

if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2023, Manfred Moitzi
# License: MIT-License
from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from typing_extensions import Self
from ezdxf.lldxf.const import SUBCLASS_MARKER, DXFStructureError
from ezdxf.lldxf.attributes import (
    DXFAttributes,
    DefSubclass,
    DXFAttr,
    group_code_mapping,
)
from .dxfentity import base_class, SubclassProcessor
from .dxfobj import DXFObject
from .factory import register_entity
from .copy import default_copy

if TYPE_CHECKING:
    from ezdxf.entities import DXFNamespace, DXFEntity
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.lldxf.tags import Tags

__all__ = ["IDBuffer", "FieldList", "LayerFilter"]

acdb_id_buffer = DefSubclass("AcDbIdBuffer", {})


@register_entity
class IDBuffer(DXFObject):
    """DXF IDBUFFER entity"""

    DXFTYPE = "IDBUFFER"
    DXFATTRIBS = DXFAttributes(base_class, acdb_id_buffer)

    def __init__(self) -> None:
        super().__init__()
        self.handles: list[str] = []

    def copy_data(self, entity: Self, copy_strategy=default_copy) -> None:
        """Copy handles"""
        assert isinstance(entity, IDBuffer)
        entity.handles = list(self.handles)

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            if len(processor.subclasses) < 2:
                raise DXFStructureError(
                    f"Missing required subclass in IDBUFFER(#{dxf.handle})"
                )
            self.load_handles(processor.subclasses[1])
        return dxf

    def load_handles(self, tags: Tags):
        self.handles = [value for code, value in tags if code == 330]

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_id_buffer.name)
        self.export_handles(tagwriter)

    def export_handles(self, tagwriter: AbstractTagWriter):
        for handle in self.handles:
            tagwriter.write_tag2(330, handle)


acdb_id_set = DefSubclass(
    "AcDbIdSet",
    {
        "flags": DXFAttr(90, default=0),  # not documented by Autodesk
    },
)
acdb_id_set_group_codes = group_code_mapping(acdb_id_set)
acdb_field_list = DefSubclass("AcDbFieldList", {})


@register_entity
class FieldList(IDBuffer):
    """DXF FIELDLIST entity"""

    DXFTYPE = "FIELDLIST"
    DXFATTRIBS = DXFAttributes(base_class, acdb_id_set, acdb_field_list)

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super(DXFObject, self).load_dxf_attribs(processor)
        if processor:
            if len(processor.subclasses) < 3:
                raise DXFStructureError(
                    f"Missing required subclass in FIELDLIST(#{dxf.handle})"
                )
            processor.fast_load_dxfattribs(dxf, acdb_id_set_group_codes, 1)
            # Load field list:
            self.load_handles(processor.subclasses[1])
        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super(DXFObject, self).export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_id_set.name)
        self.dxf.export_dxf_attribs(tagwriter, "flags")
        self.export_handles(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_field_list.name)


acdb_filter = DefSubclass("AcDbFilter", {})
acdb_layer_filter = DefSubclass("AcDbLayerFilter", {})


@register_entity
class LayerFilter(IDBuffer):
    """DXF LAYER_FILTER entity"""

    DXFTYPE = "LAYER_FILTER"
    DXFATTRIBS = DXFAttributes(base_class, acdb_filter, acdb_layer_filter)

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super(DXFObject, self).load_dxf_attribs(processor)
        if processor:
            if len(processor.subclasses) < 3:
                raise DXFStructureError(
                    f"Missing required subclass in LAYER_FILTER(#{dxf.handle})"
                )
            # Load layer filter:
            self.load_handles(processor.subclasses[2])
        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super(DXFObject, self).export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_filter.name)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_layer_filter.name)
        self.export_handles(tagwriter)



# Copyright (c) 2019-2024 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Optional, Iterator
import math
import numpy as np

from ezdxf.lldxf import validator
from ezdxf.math import (
    Vec3,
    Matrix44,
    NULLVEC,
    Z_AXIS,
    arc_segment_count,
)
from ezdxf.math.transformtools import OCSTransform, NonUniformScalingError
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
    merge_group_code_mappings,
)
from ezdxf.lldxf.const import DXF12, SUBCLASS_MARKER, DXFValueError
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import (
    DXFGraphic,
    acdb_entity,
    add_entity,
    replace_entity,
    elevation_to_z_axis,
    acdb_entity_group_codes,
)
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.entities import DXFNamespace
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.entities import Ellipse, Spline

__all__ = ["Circle"]

acdb_circle = DefSubclass(
    "AcDbCircle",
    {
        "center": DXFAttr(10, xtype=XType.point3d, default=NULLVEC),
        # AutCAD/BricsCAD: Radius is <= 0 is valid
        "radius": DXFAttr(40, default=1),
        # Elevation is a legacy feature from R11 and prior, do not use this
        # attribute, store the entity elevation in the z-axis of the vertices.
        # ezdxf does not export the elevation attribute!
        "elevation": DXFAttr(38, default=0, optional=True),
        "thickness": DXFAttr(39, default=0, optional=True),
        "extrusion": DXFAttr(
            210,
            xtype=XType.point3d,
            default=Z_AXIS,
            optional=True,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
    },
)

acdb_circle_group_codes = group_code_mapping(acdb_circle)
merged_circle_group_codes = merge_group_code_mappings(
    acdb_entity_group_codes, acdb_circle_group_codes
)


@register_entity
class Circle(DXFGraphic):
    """DXF CIRCLE entity"""

    DXFTYPE = "CIRCLE"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_circle)
    MERGED_GROUP_CODES = merged_circle_group_codes

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        """Loading interface. (internal API)"""
        # bypass DXFGraphic, loading proxy graphic is skipped!
        dxf = super(DXFGraphic, self).load_dxf_attribs(processor)
        if processor:
            processor.simple_dxfattribs_loader(dxf, self.MERGED_GROUP_CODES)
            if processor.r12:
                # Transform elevation attribute from R11 to z-axis values:
                elevation_to_z_axis(dxf, ("center",))
        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_circle.name)
        self.dxf.export_dxf_attribs(
            tagwriter, ["center", "radius", "thickness", "extrusion"]
        )

    def vertices(self, angles: Iterable[float]) -> Iterator[Vec3]:
        """Yields the vertices of the circle of all given `angles` as
        :class:`~ezdxf.math.Vec3` instances in :ref:`WCS`.

        Args:
            angles: iterable of angles in :ref:`OCS` as degrees, angle goes
                counter-clockwise around the extrusion vector, and the OCS x-axis
                defines 0-degree.

        """
        ocs = self.ocs()
        radius: float = abs(self.dxf.radius)  # AutoCAD ignores the sign too
        center = Vec3(self.dxf.center)
        for angle in angles:
            yield ocs.to_wcs(Vec3.from_deg_angle(angle, radius) + center)

    def flattening(self, sagitta: float) -> Iterator[Vec3]:
        """Approximate the circle by vertices in :ref:`WCS` as :class:`~ezdxf.math.Vec3`
        instances. The argument `sagitta`_ is the maximum distance from the center of an
        arc segment to the center of its chord. Yields a closed polygon where the start
        vertex is equal to the end  vertex!

        .. _sagitta: https://en.wikipedia.org/wiki/Sagitta_(geometry)
        """
        radius = abs(self.dxf.radius)
        if radius > 0.0:
            count = arc_segment_count(radius, math.tau, sagitta)
            yield from self.vertices(np.linspace(0.0, 360.0, count + 1))

    def transform(self, m: Matrix44) -> Circle:
        """Transform the CIRCLE entity by transformation matrix `m` inplace.
        Raises ``NonUniformScalingError()`` for non-uniform scaling.
        """
        circle = self._transform(OCSTransform(self.dxf.extrusion, m))
        self.post_transform(m)
        return circle

    def _transform(self, ocs: OCSTransform) -> Circle:
        dxf = self.dxf
        if ocs.scale_uniform:
            dxf.extrusion = ocs.new_extrusion
            dxf.center = ocs.transform_vertex(dxf.center)
            # old_ocs has a uniform scaled xy-plane, direction of radius-vector
            # in the xy-plane is not important, choose x-axis for no reason:
            dxf.radius = ocs.transform_length((dxf.radius, 0, 0))
            if dxf.hasattr("thickness"):
                # thickness vector points in the z-direction of the old_ocs,
                # thickness can be negative
                dxf.thickness = ocs.transform_thickness(dxf.thickness)
        else:
            # Caller has to catch this Exception and convert this
            # CIRCLE/ARC into an ELLIPSE.
            raise NonUniformScalingError(
                "CIRCLE/ARC does not support non uniform scaling"
            )

        return self

    def translate(self, dx: float, dy: float, dz: float) -> Circle:
        """Optimized CIRCLE/ARC translation about `dx` in x-axis, `dy` in
        y-axis and `dz` in z-axis, returns `self` (floating interface).
        """
        ocs = self.ocs()
        self.dxf.center = ocs.from_wcs(Vec3(dx, dy, dz) + ocs.to_wcs(self.dxf.center))
        # Avoid Matrix44 instantiation if not required:
        if self.is_post_transform_required:
            self.post_transform(Matrix44.translate(dx, dy, dz))
        return self

    def to_ellipse(self, replace=True) -> Ellipse:
        """Convert the CIRCLE/ARC entity to an :class:`~ezdxf.entities.Ellipse` entity.

        Adds the new ELLIPSE entity to the entity database and to the same layout as
        the source entity.

        Args:
            replace: replace (delete) source entity by ELLIPSE entity if ``True``

        """
        from ezdxf.entities import Ellipse

        layout = self.get_layout()
        if layout is None:
            raise DXFValueError("valid layout required")
        ellipse = Ellipse.from_arc(self)
        if replace:
            replace_entity(self, ellipse, layout)
        else:
            add_entity(ellipse, layout)
        return ellipse

    def to_spline(self, replace=True) -> Spline:
        """Convert the CIRCLE/ARC entity to a :class:`~ezdxf.entities.Spline` entity.

        Adds the new SPLINE entity to the entity database and to the same layout as the
        source entity.

        Args:
            replace: replace (delete) source entity by SPLINE entity if ``True``

        """
        from ezdxf.entities import Spline

        layout = self.get_layout()
        if layout is None:
            raise DXFValueError("valid layout required")
        spline = Spline.from_arc(self)
        if replace:
            replace_entity(self, spline, layout)
        else:
            add_entity(spline, layout)
        return spline



#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import NamedTuple


class FontFace(NamedTuple):
    # filename without parent directories e.g. "OpenSans-Regular.ttf"
    filename: str = ""
    family: str = "sans-serif"
    style: str = "Regular"
    weight: int = 400  # Normal - usWeightClass
    width: int = 5  # Medium(Normal) - usWidthClass

    @property
    def is_italic(self) -> bool:
        """Returns ``True`` if font face is italic."""
        return self.style.lower().find("italic") > -1

    @property
    def is_oblique(self) -> bool:
        """Returns ``True`` if font face is oblique."""
        return self.style.lower().find("oblique") > -1

    @property
    def is_bold(self) -> bool:
        """Returns ``True`` if font face weight > 400."""
        return self.weight > 400

    @property
    def weight_str(self) -> str:
        """Returns the :attr:`weight` as string e.g. "Thin", "Normal", "Bold", ..."""
        return get_weight_str(self.weight)

    @property
    def width_str(self) -> str:
        """Returns the :attr:`width` as string e.g. "Condensed", "Expanded", ..."""
        return get_width_str(self.width)

    def distance(self, font_face: FontFace) -> tuple[int, int]:
        return self.weight - font_face.weight, self.width - font_face.width


WEIGHT_STR = {
    100: "Thin",
    200: "ExtraLight",
    300: "Light",
    400: "Normal",
    500: "Medium",
    600: "SemiBold",
    700: "Bold",
    800: "ExtraBold",
    900: "Black",
}

WIDTH_STR = {
    1: "UltraCondensed",
    2: "ExtraCondensed",
    3: "Condensed",
    4: "SemiCondensed",
    5: "Medium",  # Normal
    6: "SemiExpanded",
    7: "Expanded",
    8: "ExtraExpanded",
    9: "UltraExpanded",
}


def get_weight_str(weight: int) -> str:
    """Returns the :attr:`weight` as string e.g. "Thin", "Normal", "Bold", ..."""
    key = max(min(round((weight + 1) / 100) * 100, 900), 100)
    return WEIGHT_STR[key]


def get_width_str(width: int) -> str:
    """Returns the :attr:`width` as string e.g. "Condensed", "Expanded", ..."""
    key = max(min(width, 9), 1)
    return WIDTH_STR[key]
